# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 10:38:22 2023

@author: omedeiro
"""

from time import sleep
import pyvisa
import sys
sys.path.append(r'Q:\qnnpy')
import qnnpy
import qnnpy.functions.snspd as snspd
import qnnpy.functions.functions as qf
from qnnpy.instruments.lakeshore336 import Lakeshore336
import logging

import sys, os
# sys.stdout = open("300mK.log", "w");
# sys.stderr = open("300mK.log", "w");

# https://stackoverflow.com/questions/59125493/how-to-constantly-run-python-script-in-the-background-on-windows
# print('beginning to record 300 mK')

while True:        
    try:    
        ls1 = Lakeshore336('GPIB0::13::INSTR')
    except:
        print('failed connection to Lakeshore336')
        
    try:
        stage_3K = ls1.read_temp('D4')
        sleep(2)
        stage_40K = ls1.read_temp('D2')
        sleep(2)
        stage_SORB = ls1.read_temp('D3')
        sleep(2)
        stage_He3Pot = ls1.read_temp('D1')
        sleep(2)
        print([stage_3K, stage_40K, stage_SORB, stage_He3Pot])
    except:
        print('failed to read temp')
    # ls1.read_temp('D')
    
    try:
        qf.log_data_to_database(table_name='janis_300mK_logging', stage_3K=stage_3K, stage_40K=stage_40K, stage_SORB=stage_SORB, stage_He3Pot=stage_He3Pot)
    except:
        print('failed to write to database')
        
    sleep(60)
    
    


# This file contains the class for the Keithley 2001 Multimeter
import pyvisa


class Keithley2001(object):
    """Python class for Keithley 2001 Multimeter, written by Adam McCaughan"""

    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

        # Anything else here that needs to happen on initialization

    def reset(self):
        self.write("*RST")

    def set_voltage_range(self, ran=10):
        self.write(":VOLT:RANG %0.1e" % ran)

    def set_current_range(self, ran=10e-3):
        self.write(":CURR:RANG %0.1e" % ran)

    def set_voltage_rate(
        self, nplcs=1
    ):  # specify read speed in terms of NPLCs. 1 PLC = 1/60th of a second
        self.write(":VOLT:NPLC %0.1e" % nplcs)

    def set_current_rate(
        self, nplcs=1
    ):  # specify read speed in terms of NPLCs. 1 PLC = 1/60th of a second
        self.write(":CURR:NPLC %0.1e" % nplcs)

    def read_voltage(self):
        read_str = self.query(":READ?")
        # Returns something like '+1.99919507E-01VDC,+6283.313SECS,+60584RDNG#'
        E_location = read_str.find("E")  # Finds location of E in read value
        voltage_str = read_str[0 : read_str.find("E") + 4]
        return float(voltage_str)  # Return just the first number (voltage)



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 27 16:40:17 2020

@author: omedeiro
"""

from time import sleep

import numpy as np

import qnnpy.functions.functions as qf


class Microwave:
    """Class for microwave measurements. This class handels the instrument
    configuration, measurement, plotting, logging, and saving.
    See qnnpy\examples to see this class in action"""

    def __init__(self, configuration_file):
        #######################################################################
        #       Load .yaml configuraiton file
        #######################################################################

        # qf is a package of base functions that can be used in any class.
        self.properties = qf.load_config(configuration_file)

        self.sample_name = self.properties["Save File"]["sample name"]
        self.device_name = self.properties["Save File"]["device name"]
        self.device_type = self.properties["Save File"]["device type"]

        self.R_srs = self.properties["iv_sweep"]["series_resistance"]
        self.isw = 0
        self.instrument_list = []

        #######################################################################
        # Setup instruments
        #######################################################################

        # Counter
        if self.properties.get("Counter"):
            self.instrument_list.append("Counter")
            if self.properties["Counter"]["name"]:
                if self.properties["Counter"]["name"] == "Agilent53131a":
                    from qnnpy.instruments.agilent_53131a import Agilent53131a

                    try:
                        self.counter = Agilent53131a(self.properties["Counter"]["port"])
                        # without the reset command this section will evaluate connected
                        # even though the GPIB could be wrong
                        # similary story for the other insturments
                        self.counter.reset()
                        self.counter.basic_setup()
                        # self.counter.write(':EVEN:HYST:REL 100')
                        print("COUNTER: connected")
                    except Exception:
                        print("COUNTER: failed to connect")
                else:
                    raise NameError(
                        "Invalid counter. Counter name is not " "configured"
                    )

        # Attenuator
        if self.properties.get("Attenuator"):
            self.instrument_list.append("Attenuator")

            if self.properties["Attenuator"]["name"] == "JDSHA9":
                from qnnpy.instruments.jds_ha9 import JDSHA9

                try:
                    self.attenuator = JDSHA9(self.properties["Attenuator"]["port"])
                    self.attenuator.set_beam_block(True)
                    print("ATTENUATOR: connected")
                except Exception:
                    print("ATTENUATOR: failed to connect")
            else:
                raise NameError("Invalid Attenuator. Attenuator name is not configured")

        # Scope
        if self.properties.get("Scope"):
            self.instrument_list.append("Scope")

            if self.properties["Scope"]["name"] == "LeCroy620Zi":
                from qnnpy.instruments.lecroy_620zi import LeCroy620Zi

                try:
                    self.scope = LeCroy620Zi(
                        "TCPIP::%s::INSTR" % self.properties["Scope"]["port"]
                    )
                    self.scope_channel = self.properties["Scope"]["channel"]
                    print("SCOPE: connected")
                except Exception:
                    print("SCOPE: failed to connect")
            else:
                raise NameError("Invalid Scope. Scope name is not configured")

        # Meter
        if self.properties.get("Meter"):
            self.instrument_list.append("Meter")

            if self.properties["Meter"]["name"] == "Keithley2700":
                from qnnpy.instruments.keithley_2700 import Keithley2700

                try:
                    self.meter = Keithley2700(self.properties["Meter"]["port"])
                    self.meter.reset()
                    print("METER: connected")
                except Exception:
                    print("METER: failed to connect")

            elif self.properties["Meter"]["name"] == "Keithley2400":
                # this is a source meter
                from qnnpy.instruments.keithley_2400 import Keithley2400

                try:
                    self.meter = Keithley2400(self.properties["Meter"]["port"])
                    self.meter.reset()
                    print("METER: connected")
                except Exception:
                    print("METER: failed to connect")

            elif self.properties["Meter"]["name"] == "Keithley2001":
                from qnnpy.instruments.keithley_2001 import Keithley2001

                try:
                    self.meter = Keithley2001(self.properties["Meter"]["port"])
                    self.meter.reset()
                    print("METER: connected")
                except Exception:
                    print("METER: failed to connect")
            else:
                raise NameError(
                    'Invalid Meter. Meter name: "%s" is not configured'
                    % self.properties["Meter"]["name"]
                )

        # Source
        if self.properties.get("Source"):
            self.instrument_list.append("Source")

            if self.properties["Source"]["name"] == "SIM928":
                from qnnpy.instruments.srs_sim928 import SIM928

                try:
                    self.source = SIM928(
                        self.properties["Source"]["port"],
                        self.properties["Source"]["port_alt"],
                    )
                    self.source.reset()
                    print("SOURCE: connected")
                except Exception:
                    print("SOURCE: failed to connect")

            elif self.properties["Source"]["name"] == "YokogawaGS200":
                from qnnpy.instruments.yokogawa_gs200 import YokogawaGS200

                try:
                    self.source = YokogawaGS200(self.properties["Source"]["port"])
                    # self.source.reset()
                    self.source.set_output(False)
                    self.source.set_voltage_range(5)
                    print("SOURCE: connected")
                except Exception:
                    print("SOURCE: failed to connect")

            else:
                raise NameError(
                    'Invalid Source. Source name: "%s" is not configured'
                    % self.properties["Source"]["name"]
                )

        # AWG
        if self.properties.get("AWG"):
            self.instrument_list.append("AWG")

            if self.properties["AWG"]["name"] == "Agilent33250a":
                from qnnpy.instruments.agilent_33250a import Agilent33250a

                try:
                    self.awg = Agilent33250a(self.properties["AWG"]["port"])
                    self.awg.reset()
                    print("AWG: connected")
                except Exception:
                    print("AWG: failed to connect")
            else:
                raise NameError(
                    'Invalid AWG. AWG name: "%s" is not configured'
                    % self.properties["AWG"]["name"]
                )

        # VNA
        if self.properties.get("VNA"):
            self.instrument_list.append("VNA")

            if self.properties["VNA"]["name"] == "KeysightN5224a":
                from qnnpy.instruments.keysight_n5224a import KeysightN5224a

                try:
                    self.VNA = KeysightN5224a(self.properties["VNA"]["port"])
                    # self.VNA.reset()
                    self.VNA.get_start()
                    print("VNA: connected")
                except Exception:
                    print("VNA: failed to connect")
            else:
                raise NameError(
                    'Invalid VNA. VNA name: "%s" is not configured'
                    % self.properties["VNA"]["name"]
                )

        # Temperature Controller
        if self.properties.get("Temperature"):
            self.instrument_list.append("Temperature")

            if self.properties["Temperature"]["name"] == "Cryocon350":
                from qnnpy.instruments.cryocon350 import Cryocon350

                try:
                    self.temp = Cryocon350(self.properties["Temperature"]["port"])
                    self.temp.channel = self.properties["Temperature"]["channel"]
                    self.properties["Temperature"]["initial temp"] = (
                        self.temp.read_temp(self.temp.channel)
                    )
                    print("TEMPERATURE: connected")
                except Exception:
                    print("TEMPERATURE: failed to connect")

            elif self.properties["Temperature"]["name"] == "Cryocon34":
                from qnnpy.instruments.cryocon34 import Cryocon34

                try:
                    self.temp = Cryocon34(self.properties["Temperature"]["port"])
                    self.temp.channel = self.properties["Temperature"]["channel"]
                    self.properties["Temperature"]["initial temp"] = (
                        self.temp.read_temp(self.temp.channel)
                    )
                    print("TEMPERATURE: connected")
                except Exception:
                    print("TEMPERATURE: failed to connect")

            elif self.properties["Temperature"]["name"] == "ICE":
                try:
                    self.properties["Temperature"]["initial temp"] = qf.ice_get_temp(
                        select=1
                    )
                    print("TEMPERATURE: connected")
                except Exception:
                    print("TEMPERATURE: failed to connect")

            elif self.properties["Temperature"]["name"] == "DEWAR":
                try:
                    self.properties["Temperature"]["initial temp"] = 4.2
                    print("TEMPERATURE: ~connected~ 4.2K")
                except Exception:
                    print("TEMPERATURE: failed to connect")

            else:
                raise NameError(
                    'Invalid Temperature Controller. TEMP name: "%s" is not configured'
                    % self.properties["Temperature"]["name"]
                )
        else:
            self.properties["Temperature"] = {"initial temp": "None"}
            # print('TEMPERATURE: Not Specified')


class FrequencyResponse(Microwave):
    """Setup measurement on VNA. This function will simply export the data
    using the current setup.

    """

    def run_sweep(self, save=None, plot=None):
        self.VNA.set_sweep_mode("SING")
        while self.VNA.get_sweep_mode() != "HOLD":
            sleep(0.1)

        measure = self.properties["frequency_response"]["measure"]
        measure_alt = self.properties["frequency_response"]["meas_alt"]
        for meas, meas_alt in zip(measure, measure_alt):
            print(meas)
            self.VNA.select_measurement(meas)

            self.VNA.write("CALC:FORM SMITh")

            self.if_bandwidth = self.VNA.get_if_bw()
            self.fspan = self.VNA.get_span()

            self.rf_power = self.VNA.get_power()

            if self.properties["Temperature"]["name"] == "ICE":
                self.currentTemp = qf.ice_get_temp(select=1)
            else:
                self.currentTemp = self.temp.read_temp(self.temp.channel)

            self.f, self.re, self.im = self.VNA.get_freq_real_imag()
            self.log_mag = 20 * np.log10(
                np.sqrt(np.square(self.re) + np.square(self.im))
            )
            self.phase = np.angle(self.re + self.im * 1j, deg=True)

            if plot:
                self.plot(meas_alt)
            if save:
                self.save(meas_alt)
            self.VNA.write("CALC:FORM MLOG")

        self.VNA.set_sweep_mode("CONT")

    def plot(self, meas):
        full_path = qf.save(self.properties, "frequency_response")
        qf.plot(
            self.f,
            self.phase,
            title=self.sample_name
            + " "
            + self.device_type
            + " "
            + self.device_name
            + meas
            + ": "
            + str(self.rf_power)
            + " dB",
            xlabel="Frequency (Hz)",
            ylabel="Phase (°)",
            path=full_path,
            show=True,
            linestyle="-",
            close=True,
        )

        qf.plot(
            self.f,
            self.log_mag,
            title=self.sample_name
            + " "
            + self.device_type
            + " "
            + self.device_name
            + meas
            + ": "
            + str(self.rf_power)
            + " dB",
            xlabel="Frequency (Hz)",
            ylabel="Magnitude (dB)",
            x_scale="linear",
            y_scale="linear",
            path=full_path,
            show=True,
            linestyle="-",
            close=True,
        )

    def save(self, meas):
        data_dict = {
            "freq": self.f,
            "re": self.re,
            "im": self.im,
            "log_mag": self.log_mag,
            "phase": self.phase,
            "if_bandwidth": self.if_bandwidth,
            "temp": self.currentTemp,
        }
        qf.save(
            self.properties,
            "frequency_response",
            data_dict,
            instrument_list=self.instrument_list,
            meas_txt=meas,
        )


class FrequencyResponseCurrentSweep(Microwave):
    def run_sweep_current(self, Ic_break=None, save=None, plot=None):
        self.VNA.write("CALC:FORM SMITh")
        start = self.properties["frequency_response_current_sweep"]["start"]
        stop = self.properties["frequency_response_current_sweep"]["stop"]
        steps = self.properties["frequency_response_current_sweep"]["steps"]

        currents = np.linspace(start, stop, steps)

        self.if_bandwidth = self.VNA.get_if_bw()
        self.fspan = self.VNA.get_span()
        self.rf_power = self.VNA.get_power()

        for current in currents:
            print("Bias set to %0.3f uA" % (current * 1e6))

            if self.properties["Source"]["name"] == "YokogawaGS200":
                self.source.setup_source_current()
                self.source.set_current_autorange(current)
                self.source.set_compliance_v(10)
                self.current = current * 1e6
            else:
                self.source.set_voltage(current * self.R_srs)
                self.current = current * 1e6

            self.source.set_output(True)

            if self.properties["Temperature"]["name"] == "ICE":
                self.currentTemp = qf.ice_get_temp(select=1)
            else:
                self.currentTemp = self.temp.read_temp(self.temp.channel)

            self.voltage = self.meter.read_voltage()

            self.VNA.set_sweep_mode("SING")
            while self.VNA.get_sweep_mode() != "HOLD":
                sleep(0.1)

            measure = self.properties["frequency_response"]["measure"]
            measure_alt = self.properties["frequency_response"]["meas_alt"]
            for meas, meas_alt in zip(measure, measure_alt):
                print(meas_alt)
                self.VNA.select_measurement(meas)

                self.VNA.write("CALC:FORM SMITh")

                if self.properties["Temperature"]["name"] == "ICE":
                    self.currentTemp = qf.ice_get_temp(select=1)
                else:
                    self.currentTemp = self.temp.read_temp(self.temp.channel)

                self.f, self.re, self.im = self.VNA.get_freq_real_imag()
                self.log_mag = 20 * np.log10(
                    np.sqrt(np.square(self.re) + np.square(self.im))
                )
                self.phase = np.angle(self.re + self.im * 1j, deg=True)

                if plot:
                    self.plot(meas_alt)
                if save:
                    self.save(meas_alt)
                self.VNA.write("CALC:FORM MLOG")

            self.VNA.set_sweep_mode("CONT")

        self.source.set_output(False)

    def plot(self, meas):
        full_path = qf.save(self.properties, "frequency_response_current_sweep")
        qf.plot(
            self.f,
            self.phase,
            title=self.sample_name
            + " "
            + self.device_type
            + " "
            + self.device_name
            + meas
            + ": "
            + str(round(self.current, 2))
            + " uA, "
            + str(self.rf_power)
            + " dB",
            xlabel="Frequency (Hz)",
            ylabel="Phase (°)",
            path=full_path,
            show=True,
            linestyle="-",
            close=True,
        )

        qf.plot(
            self.f,
            self.log_mag,
            title=self.sample_name
            + " "
            + self.device_type
            + " "
            + self.device_name
            + meas
            + ": "
            + str(self.rf_power)
            + " dB",
            xlabel="Frequency (Hz)",
            ylabel="Magnitude (dB)",
            x_scale="linear",
            y_scale="linear",
            path=full_path,
            show=True,
            linestyle="-",
            close=True,
        )

    def save(self, meas):
        data_dict = {
            "freq": self.f,
            "re": self.re,
            "im": self.im,
            "log_mag": self.log_mag,
            "phase": self.phase,
            "Ibias": self.current,
            "voltage": self.voltage,
            "if_bandwidth": self.if_bandwidth,
            "R_srs": self.R_srs,
            "temp": self.currentTemp,
        }
        qf.save(
            self.properties,
            "frequency_response_current_sweep",
            data_dict,
            instrument_list=self.instrument_list,
            meas_txt=meas,
        )



# -*- coding: utf-8 -*-
"""
Library for Keithley 6485 Picoammeter
Author: Dip Joti Paul
Version: 0.1
Date: 15th Feb, 2023
"""

import pyvisa


class Keithley6485(object):
    """Python class for Keithley 6485 Picoammeter"""

    """Allow 6485 to warm up for at least one hour before conducting the measurements for accuracy"""

    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 50000000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def local_key(self):
        self.write("SYST:KEY 17")

    def reset(self):
        self.write("*RST")

    def set_user_settings(self):
        self.write("SYST:AZER ON")  # enable autozero correction
        self.write("SYST:ZCH ON")
        self.write("CURR:RANG 2e-9")
        self.write("INIT")
        self.write("SYST:ZCOR:ACQ")
        self.write("SYST:ZCOR ON")
        self.write("SYST:ZCH OFF")
        self.write("CURR:RANG 20e-9")  # 2nA, 20nA, 200nA, 2uA, 20uA, 200uA, 2mA, 20mA
        self.write("CURR:NPLC 2")  # Set integration rate to 2 PLC

    def read_current(self):
        read_str = self.query(":READ?")
        print(read_str)
        # # Returns something like '+1.99919507E-01VDC,+6283.313SECS,+60584RDNG#'
        # E_location = read_str.find('E') # Finds location of E in read value
        # current_str = read_str[0 : read_str.find('E')+4]
        # return float(current_str)           # Return just the first number (current)



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 28 09:40:21 2023

@author: QNN
"""

import datetime

import numpy as np
import pyvisa


# the commonds can be found in the following link
# http://rfmw.em.keysight.com/bihelpfiles/Trueform/webhelp/US/Default.htm?lc=eng&cc=US&id=2197433
class KeysightDSOX(object):
    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands

    def read(self):
        return self.pyvisa.read()

    def read_raw(self):
        return self.pyvisa.read_raw()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def run(self):
        self.write(":RUN")

    def single(self):
        self.write(":SING")

    def stop(self):
        self.write(":STOP")

    def digitize(self):
        self.write(":DIG")

    def set_trigger_type(self, mode="EDGE"):
        """
        {EDGE | GLITch | PATTern | TV | DELay | EBURst | OR | RUNT | SHOLd | TRANsition | SBUS{1 | 2}}
        """
        self.write(":TRIG:MODE " + mode)

    def set_trigger_mode(self, trigger_mode="Normal"):
        """FAKE"""
        """
        {EDGE | GLITch | PATTern | TV | DELay | EBURst | OR | RUNT | SHOLd | TRANsition | SBUS{1 | 2}}
        """
        self.write(":TRIG:MODE EDGE")

    def set_trigger_source(self, source="CHAN1"):
        """
        {CHANnel<n> |EXTernal | LINE | WGEN | WGEN1 |WGEN2 | WMOD}
        """
        self.write(":TRIG:EDGE:SOUR " + source)

    def set_trigger_level(self, v=0.1):
        self.write(":TRIG:EDGE:LEV " + str(v))

    def set_trigger(self, source="C1", volt_level=0.1, slope="pos"):
        self.write(":TRIG:EDGE:LEV " + str(volt_level))

    def get_trigger_mode(self):
        return  # self.vbs_ask('app.Acquisition.TriggerMode')

        # Max 40e-6

    def set_timebase_scale(self, t=1e-6):
        self.write(":TIM:SCAL " + str(t))

    # test from here
    def set_timebase_mode(self, mode="MAIN"):
        """
        {MAIN | WINDow | XY |ROLL}
        """
        self.write(":TIM:MODE " + mode)

    def set_channel_bwLimit(self, bw="0", channel="1"):
        """'0' is false, '1' is true"""
        self.write(":CHAN" + str(channel) + ":BWL " + str(bw))

    def set_channel_display(self, display="1", channel="1"):
        self.write(":CHAN" + str(channel) + ":DISP " + str(display))

    def set_channel_impedance(self, impedance="ONEM", channel="1"):
        """{ONEMeg | FIFTy}"""
        self.write(":CHAN" + str(channel) + ":IMP " + str(impedance))

    def set_channel_invert(self, invert="0", channel="1"):
        self.write(":CHAN" + str(channel) + ":INV " + str(invert))

    def set_channel_offset(self, offset=0, channel="1"):
        self.write(":CHAN" + str(channel) + ":OFFS " + str(offset))

    def set_channel_scale(self, scale=0.1, channel="1"):
        self.write(":CHAN" + str(channel) + ":SCAL " + str(scale))

    def set_channel_autoScale(self, channel="1"):
        self.write(":AUT CHAN" + str(channel))

    def set_acquire_type(self, typ="HRES"):
        self.write(":ACQ:TYPE " + str(typ))

    def set_acquire_mode(self, mode="RTIM"):
        """{RTIMe | ETIMe |SEGMented}"""
        self.write(":ACQ:MODE " + str(mode))

    def set_segmented_counts(self, count=2):
        """2 to 1000"""
        self.write(":ACQ:MODE SEGM")
        self.write(":ACQ:SEGM:COUN " + str(count))

    def set_awg_arb_data(self, data, output="1"):
        self.write(":WGEN" + str(output) + ":ARB:DATA " + data)

    def set_awg_arb_data_clear(self, output="1"):
        self.write(":WGEN" + str(output) + ":ARB:DATA:CLE")

    def set_awg_freq(self, freq=1e3, output="1"):
        self.write(":WGEN" + str(output) + ":FREQ " + str(freq))

    def set_awg_pulse(self, width=10e-9, output="1"):
        self.write(":WGEN" + str(output) + ":FUNC:PULS:WIDT " + str(width))

    def set_awg_ramp(self, percent=50, output="1"):
        self.write(":WGEN" + str(output) + ":FUNC:RAMP:SYMM " + str(percent))

    def set_awg_square(self, dcycle=20, output="1"):
        self.write(":WGEN" + str(output) + ":FUNC:SQU:DCYC " + str(dcycle))

    def set_awg_output(self, val="1", output="1"):
        """1 for ON, 0 for OFF"""
        self.write(":WGEN" + str(output) + ":OUTP " + str(val))

    def set_awg_output_mode(self, mode="NORM", output="1"):
        """NORM | SING"""
        self.write(":WGEN" + str(output) + ":OUTP:MODE " + str(mode))

    def set_awg_amplitude(self, v=0.1, output=1):
        self.write(":WGEN" + str(output) + ":VOLT " + str(v))

    def set_awg_offset(self, v=0, output=1):
        self.write(":WGEN" + str(output) + ":VOLT:OFFS " + str(v))

    def set_awg_high(self, v=0.1, output=1):
        self.write(":WGEN" + str(output) + ":VOLT:HIGH " + str(v))

    def set_awg_low(self, v=0, output=1):
        self.write(":WGEN" + str(output) + ":VOLT:LOW " + str(v))

    def set_awg_phase(self, phase=0, output=1):
        self.write(":WGEN" + str(output) + ":TRAC:PHAS " + str(phase))

    def get_preamble(self):
        val = self.query(":WAV:PRE?")
        (
            wav_form,
            acq_type,
            wfmpts,
            avgcnt,
            x_increment,
            x_origin,
            x_reference,
            y_increment,
            y_origin,
            y_reference,
        ) = np.array(val.split(","), dtype=np.float32)
        return (
            wav_form,
            acq_type,
            wfmpts,
            avgcnt,
            x_increment,
            x_origin,
            x_reference,
            y_increment,
            y_origin,
            y_reference,
        )

    def save_screenshot(self, file_name=None, white_background=True):
        print("screen shot not saved. need to set up")
        if file_name is None:
            time_str = datetime.datetime.now().strftime("%Y-%m-%d %H-%M-%S")
            file_name = time_str

        return file_name

    def get_waveform(self, source="CHAN1", form="ASC", points="MAX"):
        self.write(":WAV:SOUR " + str(source))
        self.write(":WAV:FORM " + str(form))
        self.write(":WAV:POIN " + str(points))

        xorigin = float(self.query(":WAV:XOR?"))
        xinc = float(self.query(":WAV:XINC?"))
        yorigin = float(self.query(":WAV:YOR?"))
        yinc = float(self.query(":WAV:YINC?"))
        raw_data = self.query(":WAV:DATA?")

        data = np.array(raw_data[10:-1].split(","), dtype=np.float32)
        numsamples = len(data)

        x = np.array(range(numsamples)) * xinc + xorigin
        y = data - yorigin
        # y = data*yinc - yorigin
        return x, y

    def get_single_trace(self, channel="CHAN1", form="ASC", points="MAX"):
        self.write(":WAV:SOUR " + str(channel))
        self.write(":WAV:FORM " + str(form))
        self.write(":WAV:POIN " + str(points))

        xorigin = float(self.query(":WAV:XOR?"))
        xinc = float(self.query(":WAV:XINC?"))
        yorigin = float(self.query(":WAV:YOR?"))
        yinc = float(self.query(":WAV:YINC?"))
        raw_data = self.query(":WAV:DATA?")

        data = np.array(raw_data[10:-1].split(","), dtype=np.float32)
        numsamples = len(data)

        x = np.array(range(numsamples)) * xinc + xorigin
        y = data - yorigin
        # y = data*yinc - yorigin
        return x, y

    def set_measurement_clear(self):
        self.write(":MEAS:CLE")

    def set_measurement_stat_disp(self, val=1):
        self.write(":MEAS:STAT:DISP " + str(val))

    def set_measurement_stat_reset(self):
        self.write(":MEAS:STAT:RES")

    def set_measurement_tedge(self):
        self.write(":MEAS:TEDG +1")

    def get_measurement_results(self):
        results = self.query(":MEAS:RES?")
        return results

    """
      self.write(f':WAVeform:SOURce {channel}')
            self.write(f':WAVeform:FORMat {data_type}')
            waveform_data = self.oscilloscope.query_binary_values(':WAVeform:DATA?', datatype=data_type, container=np.array)
    
    """

    ":DIGitize CHANnel1"

    ":WAVeform:SOURce CHANnel1"



# -*- coding: utf-8 -*-
"""
Created on Thu May  7 10:16:58 2020

@author: omedeiro
"""

import csv
import datetime as dt
import logging
import os
import re
import sys
import time
from datetime import datetime
from enum import Enum
from time import sleep
from typing import List

import mariadb
import numpy as np
import scipy.io
import yaml
from mariadb import Connection
from matplotlib import pyplot as plt

if sys.platform == "win32":
    import win32com.client

###############################################################################
# Plotting
###############################################################################


class LogLevel(Enum):
    all_logs = 0
    debugging = 1
    warn = 2
    error = 3
    silent = 4


def plot(
    xdata,
    ydata,
    x_scale="linear",
    y_scale="linear",
    title="",
    xlabel="",
    ylabel="",
    label="",
    linestyle="o",
    path="",
    close=True,
    show=True,
):
    """
    update to using **kwargs https://book.pythontips.com/en/latest/args_and_kwargs.html

    accepts arrays or lists of arrays. Scale is the same as plt.xscale().
    If path is specified the figure will be saved to that location.
    Close and Show are true by default.

    """

    if close:
        plt.close()

    if isinstance(ydata, list) and not isinstance(xdata, list):
        if label:
            for i in np.arange(0, len(ydata), 1):
                plt.plot(xdata, ydata[i], linestyle, label=label[i])
        else:
            for i in np.arange(0, len(ydata), 1):
                plt.plot(xdata, ydata[i], linestyle)

    elif isinstance(ydata, list) and isinstance(xdata, list):
        if label:
            for i in np.arange(0, len(ydata), 1):
                plt.plot(xdata[i], ydata[i], linestyle, label=label[i])
        else:
            for i in np.arange(0, len(ydata), 1):
                plt.plot(xdata[i], ydata[i], linestyle)
    else:
        plt.plot(xdata, ydata, linestyle)

    plt.xscale(x_scale)
    plt.yscale(y_scale)

    if xlabel:
        plt.xlabel(xlabel)
    if ylabel:
        plt.ylabel(ylabel)
    if title:
        plt.title(title)
    if label:
        plt.legend()
    if not path == "":
        if isinstance(path, tuple):
            path = path[0]
        plt.savefig(path + ".png", bbox_inches="tight")
        print("File Saved:\n %s" % path)
    if show:  # show should always be after save
        plt.show()
    return plt


def plot_new(x, y, **kwargs):
    plt.plot(x, y)
    for key, value in kwargs.items():
        plt.__dict__[key](value)

    return plt


# Requires IPython for interactive shell
class LivePlotter:
    """
    Automatically updating plotter
    Requires IPython to be enabled for interactive shell
    Simpily call plot(x, y) and the plot will add your points live
    Once you're done, you can save by calling save()
    """

    # data: dict[str, (list[float],list[float],list[float])]
    data: dict

    # __subplots: dict[str, plt.Subplot]
    def __init__(
        self,
        *,
        title: str = "",
        xlabel: str = "",
        ylabel: str = "",
        legend: bool = False,
        legend_loc: str = "best",
        max_len: int = -1,
    ):
        self.data = {}
        if not plt.isinteractive():
            plt.ion()
        self.fig, self.ax = plt.subplots()
        # self.__subplots = {}
        # self.start_time: float = 0
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        self.show_legend: bool = legend
        self.legend_loc: str = legend_loc
        self.max_len = max_len
        self.def_col_idx = 0
        self.colors = ["r", "g", "b", "c", "m", "y", "k"]
        # plt.xlim(0,1)
        # plt.ylim(0,1)
        plt.draw()

    def plot(
        self,
        x: float,
        y: float,
        label: str = "",
        *,
        linestyle="solid",
        color=None,
        marker="o",
        linewidth=3,
        markercolor=None,
    ):
        # if self.start_time == 0:
        # self.start_time = time.time()
        if markercolor is None:
            markercolor = color
        if self.data.get(label) is None:
            self.data[label] = ([x], [y], self.colors[self.def_col_idx])
            self.def_col_idx += 1
            # self.data[label]=([x],[y],[time.time()-self.start_time])
        else:
            self.data[label][0].append(x)
            self.data[label][1].append(y)
            if color is None:
                color = self.data[label][2]
            if self.max_len > 1 and len(self.data[label][0]) > self.max_len:
                self.data[label][0].pop(0)
                self.data[label][1].pop(0)
            # self.data[label][2].append(time.time()-self.start_time)
        if len(self.ax.lines) >= len(self.data):
            self.ax.lines.pop(0)  # beware memory leaks and other shenanagins
        self.ax.plot(
            self.data[label][0],
            self.data[label][1],
            label=label,
            linestyle=linestyle,
            color=color,
            marker=marker,
            linewidth=linewidth,
            markerfacecolor=markercolor,
            markeredgecolor=markercolor,
        )
        self.fig.canvas.draw_idle()
        self.fig.canvas.flush_events()
        if self.show_legend:
            self.ax.legend(loc=self.legend_loc)

    def save(self, path: str = None, name: str = None, file_type: str = "jpg"):
        if path is not None:
            if "." in path:
                temp = path.rsplit(os.sep, 1)
                path = temp[0]
                if name is None:
                    name = temp[1]
            if not os.path.exists(path):
                os.makedirs(path)
            sys.path.append(path)
        if name is None:
            name: str = time.strftime(
                f"plot_%Y-%m-%d_%H-%M-%S.{file_type}", time.gmtime()
            )
        elif "." not in name:
            name = f"{name}.{file_type}"
        self.fig.savefig(f"{path}{os.sep}{name}")


###############################################################################
# Configuration
###############################################################################


def load_config(filename=None):
    """Load config accepts the full name of a .yaml file. The sample name is
    check for correct format (XXX123).
    If verified, the file is loaded and the parameters are returned.
    """
    # Check if there is a file
    parameters = {}
    if filename is None:
        raise ValueError("Please enter filename")

    # Open file
    with open(filename) as f:
        parameters = yaml.load(f, Loader=yaml.FullLoader)

    if parameters.get("Save File"):
        if parameters.get("Save File").get("sample name"):
            sample_name = parameters["Save File"]["sample name"]
            check_sample_name(sample_name)
        elif parameters.get("Save File").get("sample name 1"):
            for i in range(5):
                if parameters["Save File"].get(f"sample name {i+1}") is None:
                    break
                check_sample_name(parameters["Save File"][f"sample name {i+1}"])

    # No longer accepting path location. FIXED PATH TO NETWORK
    #    file_path=parameters['Save File']['filepath']
    #    check_file_path(file_path)

    return parameters


def check_sample_name(sample_name):
    """The sample name is required to match the XXX123 format."""
    result = re.match("^[A-Z]{3}[0-9]{3}$", sample_name)
    if not result:
        lablog_error('Invalid Sample Name. Name entered: "%s"' % sample_name)
        raise NameError("Invalid Sample Name. String must match XXX###")


def check_file_path(file_path):
    if not os.path.exists(file_path):
        try:
            os.makedirs(file_path)
        except Exception:
            lablog_error('Invalid Path. Path entered: "%s"' % file_path)
            raise NameError("Invalid Path")


###############################################################################
# Saving
###############################################################################


def save(
    parameters, measurement, data_dict={}, instrument_list=None, db=False, meas_txt=None
):
    """Save follows the typical format of defining a data dictionary (data_dict) and saving as a .mat .
    This function requires prameters from a loaded config file. The file is saved on the S:\ drive according to the configuration settings.
    If the data_dict is not included this function returns the path created from the configuration file.

    """
    if not isinstance(parameters, dict):
        raise ValueError("save accepts dict from configured .yml file")

    if "Save Root" in parameters:
        file_path = parameters["Save Root"]
    else:
        file_path = "S:\SC\Measurements"
    # Setup variables from parameters for file path
    user = parameters["User"]["name"]
    sample_name = parameters["Save File"][
        "sample name"
    ]  # this field should describe the material SPX111 or GaN_ID#20
    device_name = parameters["Save File"][
        "device name"
    ]  # this field should describe which device is being tested
    device_type = parameters["Save File"][
        "device type"
    ]  # this field should describe device type ntron, snspd, coupler, memory

    if parameters["Save File"].get("port"):
        device_type_ext = device_type + "_" + parameters["Save File"]["port"]
        port = parameters["Save File"]["port"]
    else:
        device_type_ext = device_type
        port = 1

    if parameters["Save File"].get("cell"):
        cell_name = parameters["Save File"]["cell"] + "_"
    else:
        cell_name = None

    time_str = datetime.now().strftime("%Y-%m-%d %H-%M-%S")

    """ Shorten parameter list to only include current measurement and the instruments used"""
    if instrument_list:
        new_parameters = {}
        new_parameters["User"] = parameters["User"]
        new_parameters["Save File"] = parameters["Save File"]
        new_parameters[measurement] = parameters[measurement]
        for i in range(len(instrument_list)):
            new_parameters[instrument_list[i]] = parameters[instrument_list[i]]
        parameters = new_parameters

    """ Create folder and save .mat file. """
    while os.path.exists(file_path):
        if meas_txt:
            measurement_alt = measurement + meas_txt
        else:
            measurement_alt = measurement

        file_name = (
            sample_name
            + "_"
            + measurement_alt
            + "_"
            + device_type_ext
            + "_"
            + device_name
            + "_"
            + cell_name
            + time_str
        )
        if parameters["Save File"].get("cell"):
            file_path = os.path.join(
                file_path,
                sample_name,
                device_type,
                device_name,
                measurement,
                parameters["Save File"]["cell"],
            )
        else:
            file_path = os.path.join(
                file_path, sample_name, device_type, device_name, measurement
            )
        os.makedirs(file_path, exist_ok=True)
        full_path = os.path.join(file_path, file_name)
        if data_dict:
            scipy.io.savemat(full_path + ".mat", mdict=data_dict)
            output_log(parameters, full_path)
            print("File Saved:\n %s" % full_path)
            try:
                insert_measurement_event(
                    user, measurement, sample_name, device_type, device_name, port
                )
            except Exception as e:
                print(f"Logging to qnndb failed. ({e})")

        break

    return full_path, time_str


# same as save but uses data instruments and liveplotter class
def data_saver(
    parameters: dict,
    measurement: str,
    meas_path: str = r"S:\SC\Measurements",
    data=None,
    inst=None,
    plot=None,
    file_name_append: str = "",
):
    """

    Parameters
    ----------
    parameters : dict
        parameters from loaded yaml config file.
    measurement : str
        name of the measurement: ie, iv_sweep.
    meas_path : str, optional
        root folder location for measurements to be saved to. The default is r'S:\SC\Measurements'.
    data : Data or List[Data], optional
        data class to save. The default is None.
        if a list is provided, then iteratively saves each element seperately. If multiple samples are defined in parameters, each iterative save will use the next sample
    inst : Instruments, optional
        instruments that were used. The default is None.
    plot : LivePlotter or List[LivePlotter], optional
        optionally save the live plotter which was used. The default is None.
        if a list is provided, does the same thing as data if data is a list
    Raises
    ------
    ValueError
        if there is an error in parameters.

    Returns
    -------
    full_path : str or list[str]
        full path of where data was saved.
        if multiple samples are used, an array of paths for each sample save location is given back

    """
    # for saving multiple samples
    if (
        parameters.get("Save File")
        and parameters.get("Save File").get("sample name") is None
    ):
        res: list = []
        for i in range(4):  # maximum of 4 samples can be saved at a time
            if parameters["Save File"].get(f"sample name {i+1}") is None:
                break
            d = data[i % len(data)] if isinstance(data, list) else data
            p = plot[i % len(plot)] if isinstance(plot, list) else plot
            parameters["Save File"] = (
                {} if parameters.get("Save File") is None else parameters["Save File"]
            )
            parameters["Save File"]["sample name"] = (
                parameters["Save File"][f"sample name {i+1}"]
                if parameters["Save File"].get(f"sample name {i+1}")
                else (
                    parameters.get("Save File").get("sample name")
                    if parameters.get("Save File").get("sample name")
                    else ""
                )
            )
            res.append(
                data_saver(
                    parameters,
                    measurement,
                    meas_path,
                    data=d,
                    inst=inst,
                    plot=p,
                    file_name_append=str(i),
                )
            )
        return res
    # if type(data) == list or type(plot) == list:
    if isinstance(data, list) or isinstance(plot, list):
        res: list = []
        for i in range(max(len(data), len(plot))):
            d = data[i % len(data)] if isinstance(data, list) else data
            p = plot[i % len(plot)] if isinstance(plot, list) else plot
            res.append(
                data_saver(
                    parameters,
                    measurement,
                    meas_path,
                    data=d,
                    inst=inst,
                    plot=p,
                    file_name_append=str(i),
                )
            )
        return res

    # ensure parameters is dict
    # if type(parameters) != dict:
    if isinstance(parameters, dict) is False:
        try:
            parameters = load_config(parameters)
        except Exception as e:
            raise ValueError(
                "save accepts dict from configured .yml file, try using load_config(parameters) first!"
            )
    file_path = meas_path
    # Setup variables from parameters for file path
    user = (
        parameters["User"]["name"]
        if parameters.get("User") and parameters.get("User").get("name")
        else ""
    )
    if parameters.get("Save File"):
        sample_name = (
            parameters["Save File"]["sample name"]
            if parameters.get("Save File").get("sample name")
            else ""
        )  # this field should describe the material SPX111 or GaN_ID#20
        device_name = (
            parameters["Save File"]["device name"]
            if parameters.get("Save File").get("device name")
            else ""
        )  # this field should describe which device is being tested
        device_type = (
            parameters["Save File"]["device type"]
            if parameters.get("Save File").get("device type")
            else ""
        )  # this field should describe device type ntron, snspd, coupler, memory
    else:
        sample_name, device_name, device_type = "", "", ""
    if parameters["Save File"].get("port"):
        device_type_ext = device_type + "_" + f"port{parameters['Save File']['port']}"
        port = parameters["Save File"]["port"]
    else:
        device_type_ext = device_type
        port = 1
    # Shorten parameter list to only include current measurement and the instruments used
    if inst and len(inst.instrument_list) > 0:
        new_parameters = {
            "User": parameters.get("User"),
            "Save File": parameters.get("Save File"),
            "Measurement": parameters.get(measurement),
        }
        new_parameters = {key: parameters[key] for key in inst.instrument_list}
        # for i in range(len(inst.instrument_list)):
        #     new_parameters[inst.instrument_list[i]] = parameters[inst.instrument_list[i]]
        parameters = new_parameters
    # Create folder and save .mat file
    full_path = file_path
    if os.path.exists(file_path):
        # if meas_txt:
        #     measurement_alt = measurement+meas_txt
        # else:
        #     measurement_alt = measurement
        # makes file path
        time_str = datetime.now().strftime("%Y-%m-%d %H-%M-%S")
        file_name = (
            sample_name
            + "_"
            + measurement
            + "_"
            + device_type_ext
            + "_"
            + device_name
            + "_"
            + time_str
            + (f"_({file_name_append})" if file_name_append != "" else "")
        )
        file_path = os.path.join(
            file_path, sample_name, device_type, device_name, measurement
        )
        os.makedirs(file_path, exist_ok=True)
        full_path = os.path.join(file_path, file_name)
        if data:
            # scipy.io.savemat(full_path + '.mat', mdict=data.data)
            data.save(path=f"{full_path}.mat")
            output_log(parameters, full_path)
            print("File Saved:\n %s" % full_path)
            # make "recents" shortcut
            try:
                if sys.platform == "win32":
                    recents = os.path.join(meas_path, "recents")
                    os.makedirs(recents, exist_ok=True)
                    recents_path = os.path.join(
                        recents, f"{sample_name}_measurement_{time_str}.lnk"
                    )
                    target = file_path
                    shell = win32com.client.Dispatch("WScript.Shell")
                    shortcut = shell.CreateShortCut(recents_path)
                    shortcut.Targetpath = target
                    shortcut.save()
                else:
                    print("Only windows is supported for recent measurement shortcut")
            except Exception as e:
                print(f"failed to make shortcut in for recents: {e}")
            # saving to database
            try:
                log_data_to_database(
                    "measurement_events",
                    connection=None,
                    user=user,
                    meas_type=measurement,
                    sample_name=sample_name,
                    device_type=device_type,
                    device_id=device_name,
                    port=port,
                )
                # insert_measurement_event(user, measurement, sample_name, device_type, device_name, port)
            except Exception as e:
                print(f"Logging to qnndb failed: {e}")
        if plot:
            plot.save(path=f"{full_path}.png")
    else:
        print(
            "\033[1;31;49mmeas_path does not exist, forcing a save elsewhere: \033[1;37;49m"
        )
        data.save(printloc=True)
        plot.save(name="forced_plot_save")
    return full_path


###########################################################################
# Loggging
###########################################################################


def insert_measurement_event(
    user, meas_type, sample_name, device_type, device_id, port=1
):
    """ """
    conn = database_connection()
    # Get Cursor
    cur = conn.cursor()

    sql = (
        "INSERT INTO `measurement_events` (`user`, `meas_type`, `sample_name`, `device_type`, `device_id`, `port`) VALUES ('%s', '%s', '%s', '%s', '%s', '%s')"
        % (user, meas_type, sample_name, device_type, device_id, port)
    )
    cur.execute(sql)
    conn.commit()
    return cur


def lablog_error(message):
    """
    The lablog method logs errors to 'S:\SC\ErrorLogging'
    """
    # formatter = logging.Formatter('%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
    # log = logging.getLogger('lablog')
    # if not log.handlers:
    #     handler = logging.FileHandler('S:\SC\Logging\qnn-lablog.log')
    #     handler.setFormatter(formatter)

    #     log.setLevel(logging.ERROR)
    #     log.addHandler(handler)
    #     log.propagate = False

    # log.error(message)

    timestamp = str(datetime.now()) + " "
    path = r"S:\SC\ErrorLogging\lablog_error.txt"
    file = open(path, "a")

    file.write(timestamp + message + " \n")
    file.close()


def lablog_measurement(parameters, measurement=None):
    """
    The lablog_measurement method logs the measurement history within the lab.

    """
    if isinstance(parameters, dict) is False:
        raise ValueError("log_measurement accepts dictionary from configured .yml file")
    formatter = logging.Formatter(
        "%(asctime)s %(message)s", datefmt="%m/%d/%Y %I:%M:%S %p"
    )
    lab_measurement_log = logging.getLogger("lab_measurement_log")
    if not lab_measurement_log.handlers:
        handler = logging.FileHandler(
            "S:\SC\MeasurementLogging\qnn-lablog-measurement.log"
        )
        handler.setFormatter(formatter)

        lab_measurement_log.setLevel(logging.INFO)
        lab_measurement_log.addHandler(handler)
        lab_measurement_log.propagate = False

    if measurement:
        parameters["Save File"]["measurement"] = measurement
    lab_measurement_log.info(parameters)


def output_log(parameters, path):
    """
    The output_log method logs the configuration file used for each measurement
    to the file location where that measurement is saved.

    """
    # formatter = logging.Formatter('%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')

    # output_log = logging.getLogger('output_log')
    # formatter = logging.Formatter('%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
    # handler = logging.FileHandler(path+'.txt')
    # handler.setFormatter(formatter)

    # output_log.setLevel(logging.INFO)
    # output_log.addHandler(handler)

    # output_log.info(parameters)

    file = open(path + ".txt", "w")
    file.write("\n".join("{}\t\t{}".format(k, v) for k, v in parameters.items()))
    file.close()


def database_connection(**kwargs) -> Connection:
    if kwargs is None or len(kwargs) == 0:
        try:
            with open(r"S:\SC\mariadb_conn.yml") as f:
                conn_params = yaml.load(f, Loader=yaml.FullLoader)
                conn = mariadb.connect(**conn_params)
        except mariadb.Error as e:
            print(f"Error connecting to MariaDB Platform: {e}")
            raise ConnectionError
    else:
        try:
            conn: Connection = mariadb.connect(kwargs)
        except mariadb.Error as e:
            print(f"Error connecting to MariaDB Platform: {e}")
            raise ConnectionError
    return conn


def log_data_to_database(table_name: str, connection=None, **kwargs):
    if connection is None:
        connection = database_connection()
    cur = connection.cursor()
    column_names = "`" + "`, `".join(kwargs.keys()) + "`"
    values = "'" + "', '".join([str(v) for v in kwargs.values()]) + "'"
    command = "INSERT INTO `%s` (%s) VALUES (%s)" % (table_name, column_names, values)
    cur.execute(command)
    connection.commit()
    connection.close()


def update_table(
    table_name: str, set_col: str, conditional: str = "NULL", connection=None
):
    """

    Parameters
    ----------
    table_name : str
        name of table to update.
    set_col : str
        SET sql command, for example set_col = 'thickness=0, tc=0'.
    conditional : str
        WHERE sql command, arguments without '=' will simpily get matched using '=' operator to columns in set_col.
        to update an entire column without conditional, 'ALL'
    connection : TYPE, optional
        connection to database. The default is qnndb database.
    Returns
    -------
    None.

    """
    if connection is None:
        connection = database_connection()
    command: str = "UPDATE %s SET %s" % (table_name, insert_quotes(set_col))
    if not conditional == "ALL":
        conditional = insert_quotes(conditional)
        command += " WHERE "
        conditional_operators: list[str] = [
            "=",
            ">",
            "<",
            ">=",
            "<=",
            "!=",
            "BETWEEN",
            "LIKE",
            "IN",
        ]
        if any(op in conditional for op in conditional_operators):
            command += conditional
        else:
            cols: list[str] = get_column_names(set_col)
            vals: list[str] = conditional.split(",")
            for c, v in zip(cols, vals):
                command += f"{c}={v.strip()}, "
            command = command.rstrip(", ")
    cur = connection.cursor()
    print(command)
    cur.execute(command)
    connection.commit()
    connection.close()


def get_column_names(string: str):
    """
    Helper method for update_table to get the column names when an input is formatted as 'col=val, col2=val2, col3=val3' etc
    Parameters
    ----------
    string : str
        sql command formatted as 'column=value, column=value...'

    Returns
    -------
    list of column names
    """
    res = []
    builder = ""
    build = True
    for c in string:
        if c == "=":
            res.append(builder)
            builder = ""
            build = False
        elif build:
            if not c.isspace():
                builder += c
        elif c == ",":
            build = True
    return res


def insert_quotes(string: str) -> str:
    res: str = ""
    i: int = 0
    while i < len(string):
        c = string[i]
        res += c
        if c == "=":
            is_string = False
            builder = ""
            end: int = i
            for j in range(i + 1, len(string)):
                end = j
                if string[j] == "," or (is_string and string[j] == " "):
                    end = j - 1
                    break
                if not (
                    string[j].isnumeric()
                    or string[j] == "."
                    or string[j] == "-"
                    or string[j].isspace()
                ):
                    is_string = True
                builder += string[j]
            builder = builder.strip()
            if not is_string or builder.upper() == "NULL":
                res += builder
            else:
                res += f"'{builder}'"
            i = end
        i += 1
    return res


###########################################################################
# Measurement
###########################################################################


def ice_get_temp(select=None):
    """Access ICE log to return lakeshore 336 temperature.

    If the optional input 'select' is specified (as an int between 0-4)
    the dictionary index is returned. Else all of the
    values are returned as a dictionary.

    data_dict = {'date1':date1,
                   'temp1':temp1,
                   'temp2':temp2,
                   'temp3':temp3,
                   'temp4':temp4}

    """
    now = datetime.now()
    while not os.path.exists(
        r"S:\SC\InstrumentLogging\Cryogenics\Ice\ice-log\Results\%s" % str(now)[:10]
    ):
        sleep(1)

    f = (
        r"S:\SC\InstrumentLogging\Cryogenics\Ice\ice-log\Results\%s\%s"
        % (str(now)[:10], str(now)[:10])
        + ".log"
    )

    with open(f) as f:  # Get txt from log file and split the last (most recent) entry
        last = f.readlines()[-1].split(",")

        then = datetime.strptime(last[0] + " " + last[1], "%m/%d/%Y %I:%M:%S %p")

        if now - then > dt.timedelta(minutes=1):
            temp1 = ""
            print("TEMPERATURE: ICE Logging is off")
        else:
            date1 = last[3]
            temp1 = float(last[4])  # A
            temp2 = float(last[5])
            temp3 = float(last[6])
            temp4 = float(last[7])

    data_dict = {
        "date1": date1,
        "temp1": temp1,
        "temp2": temp2,
        "temp3": temp3,
        "temp4": temp4,
    }
    data_list = [date1, temp1, temp2, temp3, temp4]
    if select:
        select_temp = data_list[select]
        return select_temp
    else:
        return data_dict


#######################################################################
#       Code testing
#######################################################################


def mock_builder(class_to_mock) -> object:
    """
    Cool class mocking method
    Takes in a class (ie: mock_builder(float)), returns a new instance
    of that class (ie: Mockfloat), with mock versions of the original's
    methods that print out when the method is called

    Parameters
    ----------
    class_to_mock : CLASS
        Class that you want to make a mock version of, such as for testing

    Returns
    -------
    Object
        Mock instance of the inputted class.

    """
    method_list: List[str] = [
        func for func in dir(class_to_mock) if callable(getattr(class_to_mock, func))
    ]
    gen_code: str = f"class Mock{class_to_mock.__name__}:"
    for m in method_list:
        if not (m.startswith("__") and not m == "__init__"):
            gen_code += f"\n\tdef {m}(*placeholder):\n\t\tprint('\033[1;33;49mMocking: \033[1;36;49mcalled \033[1;35;49m{m}()\033[1;36;49m in \033[0;35;49mMock{class_to_mock.__name__}\033[1;37;49m')\n\t\treturn None"
    exec(gen_code)
    return eval(f"Mock{class_to_mock.__name__}")


#######################################################################
#       Instrument Setup
#######################################################################
class Instruments:
    """
    Instruments general setup now supports using multiple of the same instrument.
    Currently duplicate instruments are created by naming the instrument in the
    yaml file as Source1, Source2... and are accessed using instruments(or whatever
    you named your Instruments variable).source1, instuments.source2... If you don't
    postfix your yaml instrument type with a number, it's assumed that only one
    of that instrument is used, and that instrument is accessed normally using
    inst.source (without number).

    """

    attenuator = None
    counter = None
    scope = None
    meter = None
    source = None
    awg = None
    VNA = None
    temp = None

    def __init__(self, properties: dict):
        self.attenuator = None
        self.counter = None
        self.scope = None
        self.meter = None
        self.source = None
        self.awg = None
        self.VNA = None
        self.temp = None
        self.instrument_list: List[str] = []
        self.instrument_dict: dict[str, object] = {}
        # Attenuator
        if properties.get("Attenuator"):
            self.attenuator_setup(properties)
        elif properties.get("Attenuator1"):
            for i in range(
                1, 100
            ):  # if you're using 100 or more attenuators then maybe don't use 100 attenuators? idk man
                if properties.get(f"Attenuator{i}"):
                    self.attenuator_setup(properties, i)
                else:
                    break

        # Counter
        if properties.get("Counter"):
            self.counter_setup(properties)
        elif properties.get("Counter1"):
            for i in range(1, 100):
                if properties.get(f"Counter{i}"):
                    self.counter_setup(properties, i)
                else:
                    break

        # Scope
        if properties.get("Scope"):
            self.scope_setup(properties)
        elif properties.get("Scope1"):
            for i in range(1, 100):
                if properties.get(f"Scope{i}"):
                    self.scope_setup(properties, i)
                else:
                    break

        # Meter
        if properties.get("Meter"):
            self.meter_setup(properties)
        elif properties.get("Meter1"):
            for i in range(1, 100):
                if properties.get(f"Meter{i}"):
                    self.meter_setup(properties, i)
                else:
                    break

        # Source
        if properties.get("Source"):
            self.source_setup(properties)
        elif properties.get("Source1"):
            for i in range(1, 100):
                if properties.get(f"Source{i}"):
                    self.source_setup(properties, i)
                else:
                    break

        # AWG
        if properties.get("AWG"):
            self.AWG_setup(properties)
        elif properties.get("AWG1"):
            for i in range(1, 100):
                if properties.get(f"AWG{i}"):
                    self.AWG_setup(properties, i)
                else:
                    break

        # VNA
        if properties.get("VNA"):
            self.VNA_setup(properties)
        elif properties.get("VNA1"):
            for i in range(1, 100):
                if properties.get(f"VNA{i}"):
                    self.VNA_setup(properties, i)
                else:
                    break

        # Temperature Controller
        if properties.get("Temperature"):
            self.temp_setup(properties)
        elif properties.get("Temperature1"):
            for i in range(1, 100):
                if properties.get(f"Temperature{i}"):
                    self.temp_setup(properties, i)
                else:
                    break
        else:
            properties["Temperature"] = {"initial temp": "None", "name": "None"}

        # SourceMeter
        if properties.get("Sourcemeter"):
            self.sourcemeter_setup(properties)
        elif properties.get("Sourcemeter1"):
            for i in range(1, 100):
                if properties.get(f"Sourcemeter{i}"):
                    self.sourcemeter_setup(properties, i)
                else:
                    break

    def attenuator_setup(self, properties: dict, instrument_num: int = 0):
        appender: str = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"Attenuator{appender}"
        self.instrument_list.append(inst_name)

        if properties[inst_name]["name"] == "JDSHA9":
            from qnnpy.instruments.jds_ha9 import JDSHA9

            attenuator_class = JDSHA9
        elif properties[inst_name]["name"] == "N7752A":
            from qnnpy.instruments.keysight_n7752a import N7752A

            attenuator_class = N7752A
        else:
            raise NameError("Invalid Attenuator. Attenuator name is not configured")

        try:
            attenuator = attenuator_class(properties[inst_name]["port"])
            attenuator.set_beam_block(True)
            self.instrument_dict[inst_name] = attenuator
            print(f"ATTENUATOR{appender}: connected")
        except Exception:
            print(f"ATTENUATOR{appender}: failed to connect")

        if instrument_num == 0:
            self.attenuator = attenuator
        if instrument_num == 1:
            self.attenuator1 = attenuator
        if instrument_num > 1:
            raise NotImplementedError("More than 2 attenuators not supported yet")

    def counter_setup(self, properties: dict, instrument_num: int = 0):
        appender: str = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"Counter{appender}"
        self.instrument_list.append(inst_name)

        if properties[f"Counter{appender}"]["name"] == "Agilent53131a":
            from qnnpy.instruments.agilent_53131a import Agilent53131a

            counter_class = Agilent53131a
        elif properties[f"Counter{appender}"]["name"] == "Keysight53230a":
            from qnnpy.instruments.keysight_53230a import Keysight53230a

            counter_class = Keysight53230a
        else:
            raise NameError(
                f"Invalid counter. Counter name {properties[inst_name]['name']} is not configured"
            )

        try:
            counter = counter_class(properties[inst_name]["port"])
            # without the reset command this section will evaluate connected
            # even though the GPIB could be wrong
            # similary story for the other insturments
            counter.reset()
            counter.basic_setup()
            self.instrument_dict[inst_name] = counter
            # self.counter.write(':EVEN:HYST:REL 100')
            print(f"COUNTER{appender}: connected")
        except Exception:
            print(f"COUNTER{appender}: failed to connect")
        if instrument_num == 0:
            self.counter = counter
        if instrument_num == 1:
            self.counter1 = counter
        if instrument_num > 1:
            raise NotImplementedError("More than 2 counters not supported yet")

    def scope_setup(self, properties: dict, instrument_num: int = 0):
        appender: str = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"Scope{appender}"
        self.instrument_list.append(inst_name)

        if properties[inst_name]["name"] == "LeCroy620Zi":
            from qnnpy.instruments.lecroy_620zi import LeCroy620Zi

            scopeClass = LeCroy620Zi
        elif properties[f"Scope{appender}"]["name"] == "KeysightDSOX":
            from qnnpy.instruments.keysight_dsox import KeysightDSOX

            scopeClass = KeysightDSOX
        else:
            raise NameError("Invalid Scope. Scope name is not configured")

        if properties[inst_name]["port"][0:3] == "USB":
            visa_address = properties[inst_name]["port"]
        else:
            visa_address = f"TCPIP::{properties[inst_name]['port']}::INSTR"

        try:
            scope = scopeClass(visa_address)
            print(f"SCOPE{appender}: connected")
        except Exception:
            print(f"SCOPE{appender}: failed to connect")

        if instrument_num == 0:
            self.scope = scope
            self.scope_channel = properties[inst_name]["channel"]
            self.instrument_dict[inst_name] = scope
        if instrument_num == 1:
            self.scope1 = scope
            self.scope1_channel = properties[inst_name]["channel"]
            self.instrument_dict[inst_name] = scope
        if instrument_num > 1:
            raise NotImplementedError("More than 2 scopes not supported yet")

    def meter_setup(self, properties: dict, instrument_num: int = 0):
        """
        Sets up a meter instrument based on the configuration in properties.

        Args:
            properties: A dictionary containing instrument configuration details.
            instrument_num: An integer specifying the instrument number (optional, defaults to 0).

        Raises:
            NameError: If an invalid meter name is encountered in the configuration.
        """

        appender = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"Meter{appender}"
        self.instrument_list.append(inst_name)

        # Instrument selection and import based on configuration
        meter_name = properties[inst_name]["name"]
        if meter_name == "Keithley2700":
            from qnnpy.instruments.keithley_2700 import Keithley2700

            meter_class = Keithley2700
        elif meter_name == "Keithley2400":
            from qnnpy.instruments.keithley_2400 import Keithley2400

            meter_class = Keithley2400
        elif meter_name == "Keithley2001":
            from qnnpy.instruments.keithley_2001 import Keithley2001

            meter_class = Keithley2001
        else:
            raise NameError(
                f'Invalid Meter. Meter name: "{meter_name}" is not configured'
            )

        # Instrument connection and initialization
        try:
            meter = meter_class(properties[inst_name]["port"])
            meter.reset()  # Assuming reset is a common function for all meters
            self.instrument_dict[inst_name] = meter
            print(f"METER{appender}: connected")
        except Exception as e:
            print(f"METER{appender}: failed to connect ({e})")

        # Assign meter object to class attributes based on instrument number
        if instrument_num == 0:
            self.meter = meter
        elif instrument_num == 1:
            self.meter1 = meter
        else:
            raise NotImplementedError("More than 2 meters not supported yet")

    def source_setup(self, properties: dict, instrument_num: int = 0):
        """
        Sets up a source instrument based on the configuration in properties.

        Args:
            properties: A dictionary containing instrument configuration details.
            instrument_num: An integer specifying the instrument number (optional, defaults to 0).

        Raises:
            NameError: If an invalid source name is encountered in the configuration.
        """

        appender = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"Source{appender}"
        self.instrument_list.append(inst_name)

        # Instrument selection and import based on configuration
        source_name = properties[inst_name]["name"]
        if source_name == "SIM928":
            from qnnpy.instruments.srs_sim928 import SIM928

            source_class = SIM928
        elif source_name == "YokogawaGS200":
            from qnnpy.instruments.yokogawa_gs200 import YokogawaGS200

            source_class = YokogawaGS200
        elif source_name == "Keithley2400":
            from qnnpy.instruments.keithley_2400 import Keithley2400

            source_class = Keithley2400
        else:
            raise NameError(
                f'Invalid Source. Source name: "{source_name}" is not configured'
            )

        # Instrument connection and initialization
        try:
            source = source_class(
                properties[inst_name]["port"],
                properties[inst_name].get("port_alt", None),  # Handle optional port_alt
            )
            source.reset()  # Assuming reset is a common function for all sources
            source.set_output(False)  # Assuming this is a common configuration step
            self.instrument_dict[inst_name] = source
            print(f"SOURCE{appender}: connected")
        except Exception as e:
            print(f"SOURCE{appender}: failed to connect ({e})")

        # Assign source object to class attributes based on instrument number
        if instrument_num == 0:
            self.source = source
        elif instrument_num == 1:
            self.source1 = source
        else:
            raise NotImplementedError("More than 2 sources not supported yet")

    def sourcemeter_setup(self, properties: dict, instrument_num: int = 0):
        """
        Sets up a sourcemeter instrument based on the configuration in properties.

        Args:
            properties: A dictionary containing instrument configuration details.
            instrument_num: An integer specifying the instrument number (optional, defaults to 0).

        Raises:
            NameError: If an invalid sourcemeter name is encountered in the configuration.
        """

        appender = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"Sourcemeter{appender}"
        self.instrument_list.append(inst_name)

        # Instrument selection and import based on configuration
        sourcemeter_name = properties[inst_name]["name"]
        if sourcemeter_name == "KeysightB2912a":
            from qnnpy.instruments.keysight_b2912a import KeysightB2912a

            sourcemeter_class = KeysightB2912a
        else:
            raise NameError(
                f'Invalid Sourcemeter. Sourcemeter name: "{sourcemeter_name}" is not configured'
            )

        # Instrument connection and initialization
        try:
            sourcemeter = sourcemeter_class(properties[inst_name]["port"])
            sourcemeter.reset()  # Assuming reset is a common function for all sourcemeters
            sourcemeter.set_output(
                False
            )  # Assuming this is a common configuration step
            self.instrument_dict[inst_name] = sourcemeter
            print(f"SOURCEMETER{appender}: connected")
        except Exception as e:
            print(f"SOURCEMETER{appender}: failed to connect ({e})")

        # Assign sourcemeter object to class attributes based on instrument number
        if instrument_num == 0:
            self.sourcemeter = sourcemeter
        elif instrument_num == 1:
            self.sourcemeter1 = sourcemeter
        else:
            raise NotImplementedError("More than 2 sourcemeters not supported yet")

    def AWG_setup(self, properties: dict, instrument_num: int = 0):
        """
        Sets up an AWG instrument based on the configuration in properties.

        Args:
            properties: A dictionary containing instrument configuration details.
            instrument_num: An integer specifying the instrument number (optional, defaults to 0).

        Raises:
            NameError: If an invalid AWG name is encountered in the configuration.
            NotImplementedError: If more than 2 AWGs are requested.
        """

        appender = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"AWG{appender}"
        self.instrument_list.append(inst_name)

        # Instrument selection and import based on configuration
        awg_name = properties[inst_name]["name"]
        if awg_name == "Agilent33250a":
            from qnnpy.instruments.agilent_33250a import Agilent33250a

            awg_class = Agilent33250a
        elif awg_name == "Agilent33600a":
            from qnnpy.instruments.agilent_33600a import Agilent33600a

            awg_class = Agilent33600a
        else:
            raise NameError(
                f"Invalid AWG. AWG name: {properties[inst_name]['name']} is not configured"
            )

        # Instrument connection and initialization
        try:
            awg = awg_class(properties[inst_name]["port"])
            awg.beep()  # Assuming beep is a common function for all AWGs
            print(f"AWG{appender}: connected")
            self.instrument_dict[inst_name] = awg
            if instrument_num == 0:
                self.awg = awg
            elif instrument_num == 1:
                self.awg1 = awg
            else:
                raise NotImplementedError("More than 2 AWGs not supported yet")
        except Exception as e:
            print(f"AWG{appender}: failed to connect ({e})")

    # VNA
    def VNA_setup(self, properties: dict, instrument_num: int = 0):
        """
        Sets up a VNA instrument based on the configuration in properties.

        Args:
            properties: A dictionary containing instrument configuration details.
            instrument_num: An integer specifying the instrument number (optional, defaults to 0).

        Raises:
            NameError: If an invalid VNA name is encountered in the configuration.
        """

        appender = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"VNA{appender}"
        self.instrument_list.append(inst_name)

        # Instrument selection and import based on configuration
        vna_name = properties[inst_name]["name"]
        if vna_name == "KeysightN5224a":
            from qnnpy.instruments.keysight_n5224a import KeysightN5224a

            vna_class = KeysightN5224a
        else:
            raise NameError(
                f"Invalid VNA. VNA name: {properties[inst_name]['name']} is not configured"
            )

        # Instrument connection and initialization
        try:
            vna = vna_class(properties[inst_name]["port"])
            # Assuming reset is a common function for all VNAs, uncomment if needed
            # vna.reset()
            self.instrument_dict[inst_name] = vna
            print(f"VNA{appender}: connected")
        except Exception as e:
            print(f"VNA{appender}: failed to connect ({e})")

        if instrument_num == 0:
            self.VNA = vna
        if instrument_num > 0:
            raise NotImplementedError("More than 1 VNA not supported yet")

    # Temperature Controller
    def temp_setup(self, properties: dict, instrument_num: int = 0):
        appender: str = str(instrument_num)
        if instrument_num == 0:
            appender = ""
        inst_name = f"Temperature{appender}"
        self.instrument_list.append(inst_name)

        if properties[inst_name]["name"] == "Lakeshore336":
            from qnnpy.instruments.lakeshore336 import Lakeshore336

            temp_class = Lakeshore336
        elif properties[inst_name]["name"] == "Cryocon34":
            from qnnpy.instruments.cryocon34 import Cryocon34

            temp_class = Cryocon34
        elif properties[inst_name]["name"] == "ICE":
            temp_class = None
        elif properties[inst_name]["name"] == "DEWAR":
            temp_class = None
        else:
            raise NameError(
                'Invalid Temperature Controller. TEMP name: "%s" is not configured'
                % properties[inst_name]["name"]
            )

        if temp_class is not None:
            try:
                temp = temp_class(properties[inst_name]["port"])
                temp.channel = properties[inst_name]["channel"]
                properties[inst_name]["initial temp"] = temp.read_temp(
                    temp.channel
                )  # Assuming read_temp is a common function for all temperature controllers
                self.instrument_dict[inst_name] = temp
                print(
                    "TEMPERATURE"
                    + appender
                    + ": connected | "
                    + str(properties[inst_name]["initial temp"])
                )
            except Exception as e:
                properties[inst_name]["initial temp"] = 0
                print(f"TEMPERATURE{appender} failed to connect with message ({e})")
        else:
            if properties[inst_name]["name"] == "ICE":
                temp = None
                try:
                    properties["Temperature" + appender]["initial temp"] = ice_get_temp(
                        select=1
                    )
                    print(
                        "TEMPERATURE"
                        + appender
                        + ": connected T="
                        + str(ice_get_temp(select=1))
                    )
                except Exception:
                    properties["Temperature" + appender]["initial temp"] = 0
                    print("TEMPERATURE" + appender + ": failed to connect")
            if properties[inst_name]["name"] == "DEWAR":
                temp = None
                try:
                    properties["Temperature" + appender]["initial temp"] = 4.2
                    print("TEMPERATURE" + appender + ": ~connected~ 4.2K")
                except Exception:
                    properties["Temperature" + appender]["initial temp"] = 0
                    print("TEMPERATURE" + appender + ": failed to connect")

        if instrument_num == 0:
            self.temp = temp
        if instrument_num > 0:
            raise NotImplementedError(
                "More than 1 temperature controller not supported yet"
            )


#######################################################################
#       Temporary Data Storage and Cache-ing
#######################################################################


class Data:
    """
    The data class is used to store and save any collected data
    If no default file save location is provided, one will automatically be generated

    NOTE that autosaving does not work with .mat files due to how the binary
    data in a .mat file is stored. the only way to append to a mat file is
    to read the whole file into a python dictionary, modify, then re-write
    the entire thing, which defeats the purpose of "autosaving" to minimize
    memory usage.

    Parameters
    ----------
    autosave : bool, optional
        When enabled, periodically empties out Data and auto-saves it to the file location provided. The default is False.
        Note: If using autosave, remember to still call save() at the end to store any data in the current save_increment that hasn't been transferred yet!
    save_increment : int, optional
        How often to autosave whenever store() is called. The default is every 128th time store() is called.
    path : str, optional
        file path to save to. automatically sets up folders if full path doesn't exist. The default is None.
    name : str, optional
        file name to save to. if a name is already provided in path, it is overridden by this. The default is None.
    file_type : str, optional
        file type to save to. The default is 'csv'.
    preserve_pos_order : bool, optional
        if store(v1=1,v2=2) then store(v2=3, v3=4) is called, by default v1
        and v4 will be compressed into the first line, while v2 will appear
        on lines 1 and 2. Enabling preserve_pos_order will create empty
        columns to fix this ordering. The default is False.
    connection : mariadb.connection, optional
        If you want to auto-log data to a database, then you can set a connection here.
        Just remember to run connection.close() after you're done!
    table_name : str, optional
        database table name
    logtime: bool, optional
        logs the time in the data dict along with other variables whenever store() is called
    Returns
    -------
    None.

    """

    # data: dict[str,list[object]]
    data: dict
    numcalls: int  # number of times store is called, reset to 0 when empty() is called

    # save_increment = how often to save csv when calling store, every time (1), every other time (2)
    def __init__(
        self,
        *,
        autosave: bool = False,
        save_increment: int = 128,
        path: str = None,
        name: str = None,
        file_type: str = "csv",
        preserve_pos_order: bool = False,
        table_name: str = None,
        connection=None,
        logtime=False,
    ):
        self.data = {}
        self.numcalls = 0
        self.preserve_pos = preserve_pos_order
        self.connection = None
        self.connectionattempts = 0
        if path is not None:
            if "." in path:
                temp = path.rsplit(os.sep, 1)
                path = temp[0]
                if name is None:
                    name = temp[1]
            if not os.path.exists(path):
                os.makedirs(path)
        if name is None:
            name: str = time.strftime(
                f"data_%Y-%m-%d_%H-%M-%S.{file_type}", time.gmtime()
            )
        elif "." not in name:
            name = f"{name}.{file_type}"
        if path is None:
            self.save_loc = name
        else:
            self.save_loc = f"{path}{os.sep}{name}"
        # print(self.save_loc)
        self.autosave = autosave
        if autosave:
            # self.close_csv()
            self.save_increment = save_increment
            self.save_increment_counter = 0
        if connection is not None:
            cursor = connection.cursor()
            cursor.execute(
                """
                SELECT COUNT(*)
                FROM information_schema.tables
                WHERE table_name = '{0}'
                """.format(table_name.replace("'", "''"))
            )
            if not cursor.fetchone()[0] == 1:
                print(f"Table {table_name} does not exist.")
                connection.close()
            else:
                self.connection = connection
                self.dbtable_name = table_name
        self.logtime = logtime

    def store(self, **kwargs):
        if self.logtime:
            kwargs["time"] = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        for key in kwargs:
            if self.data.get(key) is not None:
                self.data[key].append(kwargs[key])
            else:
                if self.preserve_pos and self.numcalls > 0:
                    self.data[key] = [""] * self.numcalls
                    self.data[key].append(kwargs[key])
                else:
                    self.data[key] = [kwargs[key]]
                exec(f"self.{key}=self.data['{key}']")
        if self.connection is not None:
            try:
                log_data_to_database(self.dbtable_name, self.connection, **kwargs)
                self.connectionattempts = 0
            except Exception as e:
                self.connectionattempts += 1
                print(f"Data failed to log to database: {e}")
                if (
                    self.connectionattempts > 10
                ):  # times out connection after 10 failed attempts
                    print(
                        "\033[1;31;49mDatabase connection has timed out, closing connection...\033[1;37;49m"
                    )
                    try:
                        self.connection.close()
                    except Exception:
                        pass
                    self.connection = None
        if self.preserve_pos and len(kwargs) < len(self.data):
            for key in self.data:
                if kwargs.get(key) is None:
                    self.data[key].append("")
        if self.autosave:
            self.save_increment_counter += 1
            if self.save_increment_counter >= self.save_increment:
                self.save(path=self.save_loc)
                if self.connection is not None:
                    try:
                        self.connection.commit()
                        self.connectionattempts = 0
                    except Exception as e:
                        self.connectionattempts += 1
                        print(f"Data failed to commit to database: {e}")
                        if (
                            self.connectionattempts > 10
                        ):  # times out connection after 10 failed attempts
                            print(
                                "\033[1;31;49mDatabase connection has timed out, closing connection...\033[1;37;49m"
                            )
                            try:
                                self.connection.close()
                            except Exception:
                                pass
                            self.connection = None
                self.empty()
                self.save_increment_counter = 0
        self.numcalls += 1

    def get(self, key: str) -> List[object]:
        return self.data[key]

    def last(self, key: str) -> object:
        return self.get(key)[-1]

    def empty(self):
        """
        empties out the values stored in the data dict, but retains any
        dictionary keys

        Returns
        -------
        None.

        """
        for key in self.data:
            self.data.get(key).clear()
        self.numcalls = 0

    def save(
        self,
        path: str = None,
        name: str = None,
        file_type: str = "csv",
        override: bool = False,
        printloc: bool = False,
    ):
        """
        saves the current contents of the data class to a file.
        if a default save location is provided in initialization, arguments
        provided here will override defaults, UNLESS no arguments are provided,
        (excluding override argument), in which case defaults will still be used

        note that if autosaving is enabled, then the data will periodically
        get cleared to save memory, which means not all data collected will be
        included in newer files created by save(), only the autosave file will
        include all data.

        also note: saving once to a .mat file works just fine, but attempting to
        append to a .mat file DOES NOT WORK. This also means that autosave does
        not work with .mat files.

        Parameters
        ----------
        path : str, optional
            file path to use. automatically sets up folders if full path doesn't exist. The default is None.
        name : str, optional
            file name to use. if a name is already provided in path, it is overridden by this. The default is None.
        file_type : str, optional
            file type. The default is 'csv'.
        override : bool, optional
            if to override previous file if it already exists. The default is False.
        Returns
        -------
        None
        """
        if path is None and name is None and file_type == "csv":
            path = self.save_loc
        if path is not None and os.sep not in path:
            name = path
            path = None
        if path is not None:
            if "." in path:
                temp = path.rsplit(os.sep, 1)
                path = temp[0]
                if name is None:
                    name = temp[1]
            if not os.path.exists(path):
                os.makedirs(path)
            # sys.path.append(path)
        if name is None:
            name: str = time.strftime(
                f"data_%Y-%m-%d_%H-%M-%S.{file_type}", time.gmtime()
            )
        elif "." not in name:
            name = f"{name}.{file_type}"
        if path is None:
            path = ""
        try:
            mode: str = "w"
            if os.path.exists(f"{path}{os.sep}{name}") and not override:
                mode = "a"
            if mode == "a" and name.rsplit(".", 1)[1] == "mat":
                mode = "ab"
            # print(mode + " " + str(os.path.exists(f"{path}{os.sep}{name}")) + " " + f"{path}{os.sep}{name}")
            with open(f"{path}{os.sep}{name}", mode) as f:
                if name.rsplit(".", 1)[1] == "mat":
                    # print(self.data)
                    if mode == "ab":
                        scipy.io.savemat(f, mdict=self.data)
                    else:
                        scipy.io.savemat(f"{path}{os.sep}{name}", mdict=self.data)
                    return
                writer = csv.writer(f)
                if mode == "w":
                    writer.writerow(self.data.keys())
                writer.writerows(zip(*self.data.values()))
            if printloc:
                print(f"{path}{os.sep}{name}")
        except IOError as e:
            # even if initial save attempt fails, will try to force store data in a temporary file at the script location
            print(
                f"\033[1;31;49mI/O error: {e}, attempting to force data save... \033[1;37;49m"
            )
            try:
                with open("forced_data_save.csv", "a") as f:
                    writer = csv.writer(f)
                    writer.writerow(self.data.keys())
                    writer.writerows(zip(*self.data.values()))
                print("data saved at *this user*\\forced_data_save.csv")
            except Exception as e:
                print(
                    f"Backup failed somehow, if you're seeing this things are really messed up: {e}"
                )



# -*- coding: utf-8 -*-
"""
Library for Yokogawa GS-200 DC Voltage/Current Source
Author: Luca Camellini
Version: 0.2
Date: 17th July, 2018

Inspired by amcc's Keithley2400 library

"""

import pyvisa


class YokogawaGS200(object):
    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def reset(self):
        self.write("*RST")

    def setup_source_current(self):
        self.write("*RST")
        # Set current source mode
        self.write(":SOUR:FUNC CURR")
        self.write(":SOUR:LEV:AUTO 0E-6")
        # Enable measurement mode
        self.write(":SENS:REM OFF")  # 2 wire
        # self.write(':SENS ON')

    def setup_source_voltage(self):
        self.write("*RST")
        # Set voltage source mode
        self.write(":SOUR:FUNC VOLT")
        self.write(":SOUR:LEV:AUTO 0E-3")
        # Enable measurement mode
        self.write(":SENS:REM OFF")  # 2 wire
        # self.write(':SENS ON')

    # def setup_4W_source_I_read_V(self):
    #     self.write('*RST')
    #     #Set current source mode
    #     self.write(':SOUR:FUNC CURR')
    #     self.write(':SOUR:LEV:AUTO 0E-6')
    #     #Enable measurement mode
    #     self.write(':SENS:REM ON') #4 wire
    #     self.write(':SENS ON')

    #    def set_measurement_time(self, integration_time = 1):
    #        # Set integration time in terms of power line cycle. 1 = 1/60th of a second
    #        # Min = 1, Default = 1, Max = 25
    #        self.write(':SENS :NPLC %i' %integration_time)

    # Need setup_read_volt first
    #    def read_volt(self):
    #        voltage = self.query(':MEAS?')
    #        return voltage #returns real number - see 13-17
    #
    #    #Need setup_read_curr first
    #    def read_curr(self):
    #        current = self.query(':MEAS?')
    #        return current #returns real number - see 13-17

    def set_current_range(self, ran=1):
        # Remember: Yokogawa has no auto-range feature!
        # 1 -> 1mA
        # 2 -> 10mA
        # 3 -> 100mA
        # 4 -> 200mA
        # You can also let the instrument choose the range for the current value you apply
        # Use set_current_autorange
        if ran == 1:
            self.write(":SOUR:RANG 1E-3")
        elif ran == 2:
            self.write(":SOUR:RANG 10E-3")
        elif ran == 3:
            self.write(":SOUR:RANG 100E-3")
        elif ran == 4:
            self.write(":SOUR:RANG 200E-3")

    def set_voltage_range(self, ran=1):
        # Remember: Yokogawa has no auto-range feature!
        # 1 -> 10mV
        # 2 -> 100mV
        # 3 -> 1V
        # 4 -> 10V
        # 5 -> 30V
        # You can also let the instrument choose the range for the current value you apply
        # Use set_voltage_autorange
        if ran == 1:
            self.write(":SOUR:RANG 10E-3")
        elif ran == 2:
            self.write(":SOUR:RANG 100E-3")
        elif ran == 3:
            self.write(":SOUR:RANG 1E+0")
        elif ran == 4:
            self.write(":SOUR:RANG 10E+0")
        elif ran == 5:
            self.write(":SOUR:RANG 30E+0")

    def set_current(self, current=0e-6):
        self.write(":SOUR:LEV %0.4e" % current)  # Set current level

    def set_voltage(self, voltage=0e-6):
        self.write(":SOUR:LEV %0.4e" % voltage)  # Set voltage level

    def set_guard(self, guard=False):
        # Guard terminal is used to reduce common ground noise - see page 4-6 of the manual
        if guard is True:
            self.write(":SENS:GUAR ON")
        else:
            self.write(":SENS:GUAR OFF")

    def set_current_autorange(self, current=0e-6):
        self.write(":SOUR:LEV:AUTO %0.4e" % current)  # Set current level

    def set_voltage_autorange(self, voltage=0e-6):
        self.write(":SOUR:LEV:AUTO %0.4e" % voltage)  # Set voltage level

    def set_compliance_i(self, compliance_i=1e-3):
        # Limit value for current when used in voltage source mode
        # from 1 mA to 200 mA
        self.write(":SOUR:PROT:CURR %0.3e" % compliance_i)

    def set_compliance_v(self, compliance_v=1):
        # Limit value for voltage when used in current source mode
        # from 1 V to 30 V
        self.write(":SOUR:PROT:VOLT %0.3e" % compliance_v)

    def read_voltage(self):
        read_str = self.query(":READ?")
        # Returns something like '+1.99919507E-01VDC,+6283.313SECS,+60584RDNG#'
        E_location = read_str.find("E")  # Finds location of E in read value
        voltage_str = read_str[0 : read_str.find("E") + 4]
        return float(voltage_str)  # Return just the first number (voltage)

    def set_output(self, output=False):
        if output is True:
            self.write(":OUTP ON")
        if output is False:
            self.write(":OUTP OFF")



# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 10:38:22 2023

@author: omedeiro
"""

from time import sleep
import pyvisa
import sys
sys.path.append(r'Q:\qnnpy')
import qnnpy
import qnnpy.functions.snspd as snspd
import qnnpy.functions.functions as qf
from qnnpy.instruments.lakeshore336 import Lakeshore336

# https://stackoverflow.com/questions/59125493/how-to-constantly-run-python-script-in-the-background-on-windows


while True:       
    
    try:
        ls1 = Lakeshore336('GPIB0::12::INSTR')
        ls2 = Lakeshore336('GPIB0::13::INSTR')
    except:
        print('failed connection to Lakeshore336')
    try:
        T_sample_stage = ls1.read_temp('A')
        T_4k_stage = ls1.read_temp('B')
        T_sample_holder = ls1.read_temp('C')
        # ls1.read_temp('D')
        
        T_rad_shield = ls2.read_temp('A')
        T_second_shield = ls2.read_temp('B')
    except:
        print('failed to read temp')
    try:
        qf.log_data_to_database(table_name='probe_station_logging', sample_stage=T_sample_stage, fourK_stage=T_4k_stage, sample_holder=T_sample_holder, radiation_shield=T_rad_shield, second_shield=T_second_shield)
    except:
        print('failed to write to database')
        
    sleep(1)
    


# -*- coding: utf-8 -*-
"""
Created on Thu Mar 31 22:19:54 2022

@author: omedeiro
"""

import pyvisa


class Tektronix5014(object):
    """"""

    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def set_amplitude(self, v=0.02, chan=1):
        self.write("SOUR%s:VOLT %s" % (chan, v))

    def set_high(self, v=0.02, chan=1):
        self.write("SOUR%s:VOLT %s" % (chan, v))
        self.write("SOUR%s:VOLT:OFFS %s" % (chan, v / 2))



import pyvisa


class HP8157A(object):
    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def set_attenuation_db(self, attenuation_db=10):
        self.write(("ATT %0.1f dB" % attenuation_db))

    def set_beam_block(self, beam_block=True):
        self.write(("D %0.0f" % beam_block))


# att = JDSHA9('GPIB0::10')



import pyvisa


class AgilentE5061B(object):
    """Python class for Agilent E5061B Network Analyzer, written by Adam McCaughan"""

    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def reset(self):
        self.write("*RST")
        self.timeout = 1

    def get_channel_data(self, channel=1):
        self.write(":FORM:DATA ASC")
        m = self.ask_for_values(":CALC1:DATA:FDAT?")
        m = m[::2]
        f = self.ask_for_values(":SENS1:FREQ:DATA?")
        return f, m


# na = AgilentE5061B('GPIB0::17')
# def freq_range(self, f_start = 0.1e9, f_stop = 1.0e9, num_pts = 400):

#     self.write('POIN %0.0i;' % num_pts)
#     self.write('STAR %0.6e;' % f_start)
#     self.write('STOP %0.6e;' % f_stop)

# def fixed_freq(self, f = 10e6):
#     self.write('CWFREQ%0.6eHZ;' % f)

# def power(self, power = -20):
#     if (power < -60) or (power > 0):
#         print 'Out of range value'; return
#     if   power >=  -5:  power_range = '01' # Only available below 26 GHz
#     elif power >= -20:  power_range = '02'
#     elif power >= -35:  power_range = '05'
#     elif power >= -50:  power_range = '08'
#     elif power >= -60:  power_range = '10'

#     self.write('PRAN%s' % power_range)
#     self.write('POWE %0.0d' % power)  # Sets power (in dBm)

# def s_mode(self, s_mode = 'S11'):
#     self.write('%s;' % s_mode)

# def format_polar(self):
#     self.write('POLA') # Set to polar coordinates

# def format_logarithmic(self):
#     self.write('LOGM') # Set to polar coordinates

# def run_sweep_ri(self):
#     """ Runs a sweep using whatever settings are currently on the NA and returns the real
#     and imaginary components of each data point """
#     self.format_polar() # Set to polar coordinates
#     f_start = float(self.query('STAR?;'))
#     f_span = float(self.query('SPAN?;'))
#     f_stop = float(self.query('STOP?;'))
#     num_pts = int(float(self.query('POIN?;')))

#     print 'Sweeping from %0.0d to %0.0d, with %0.0d points' % (f_start/1e6, f_stop/1e6, num_pts)
#     # FIXME: Waiting loop here
#     # while (self.no bits are waiting):
#     #     sleep(1)
#     #     print 'Operation not yet complete, waiting 1s'

#     self.write('FORM4;') # Make the data output in ASCII
#     data = self.ask_for_values('OUTPFORM;')

#     f = np.linspace(f_start, f_stop, num_pts)
#     R = data[::2] # Every other element starting with element 0
#     I = data[1::2]
#     return f, R, I



import pyvisa


class Keithley2400(object):
    """Python class for Keithley 2400 Sourcemeter, written by Adam McCaughan"""

    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands
        self.name = "Keithley2400"

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def reset(self):
        self.write("*RST")

    def local_key(self):
        self.write("SYST:KEY 23")

    def setup_read_volt(self):
        self.write("*RST")
        self.write(":SOUR:FUNC CURR")
        self.write(":SOUR:CURR:LEVEL 0E-6")
        self.write('SENS:FUNC "VOLT"')

    def setup_4W_source_I_read_V(self, current="1000"):
        """current level in microamps"""
        self.write("*RST")
        self.write(":SOUR:FUNC CURR")  # Set operation mode to: source current
        print(":SOUR:CURR:LEVEL " + current + "E-6")
        self.write(":SOUR:CURR:LEVEL " + current + "E-6")
        self.write(
            ":SYST:RSEN 1"
        )  # Turn on "Remote Sensing" aka 4-wire measurement mode
        self.write('SENS:FUNC "VOLT", "CURR"')  # Have it output

    def setup_2W_source_I_read_V(self):
        self.write("*RST")
        self.write(":SOUR:FUNC CURR")  # Set operation mode to: source current
        self.write(":SOUR:CURR:LEVEL 0E-6")  # Set current level to 0 uA
        self.write(
            ":SYST:RSEN 0"
        )  # Turn off "Remote Sensing" aka 4-wire measurement mode
        self.write('SENS:FUNC "VOLT", "CURR"')  # Have it output

    def setup_2W_source_V_read_I(self):
        self.write("*RST")
        self.write(":SOUR:FUNC VOLT")  # Set operation mode to: source voltage
        self.write(":SOUR:VOLT:LEVEL 0E-3")  # Set voltage level to 0 mV
        self.write(
            ":SYST:RSEN 0"
        )  # Turn off "Remote Sensing" aka 4-wire measurement mode
        self.write('SENS:FUNC "VOLT", "CURR"')  # Have it output

    def set_output(self, output=False):
        if output is True:
            self.write(":OUTP ON")
        if output is False:
            self.write(":OUTP OFF")

    def set_measurement_time(self, plc_cycles=1.0):
        """plc_cycles Sets integration time Keithley.  Each cycle corresponds to
        1/60th of a second.  Default is 1.0, Max is 10.0. Min is 0.01.  See Keithley
        manual p18-70 for more details"""
        self.write(":NPLCycles %0.2f" % plc_cycles)

    def disable_remote(self):
        """Simulates the pressing of the "LOCAL" button on the Keithley
        which will take the keithley out of remote mode"""
        self.write(":SYST:KEY 23")

    def set_compliance_i(self, compliance_i=10e-6):
        self.write(":SENS:CURR:PROT %0.3e" % compliance_i)

    def set_compliance_v(self, compliance_v=10e-6):
        self.write(":SENS:VOLT:PROT %0.3e" % compliance_v)

    def set_current(self, current=0e-6):
        self.write(":SOUR:CURR:LEVEL %0.4e" % current)  # Set current level

    def set_voltage(self, voltage=0e-6):
        self.write(":SOUR:VOLT:LEVEL %0.4e" % voltage)  # Set current level

    def read_voltage_and_current(self):
        read_str = self.query(":READ?")
        # See page 18-51 of manual, returns: voltage, current, resistance, timestamp, status info
        # Returns something like '5.275894E-05,-1.508318E-06,+9.910000E+37,+2.562604E+03,+3.994000E+04'
        data = read_str.split(",")
        # print(read_str[0])
        voltage = float(data[0])
        # voltage= float(data[0][:-4])
        # voltage= data
        # print(data)
        # current = float(data[1])
        current = 0
        return voltage, current

    def read_current(self, current=0e-6):
        voltage, current = self.read_voltage_and_current()
        return current

    def read_voltage(self):
        voltage, current = self.read_voltage_and_current()
        return voltage

    def read_resistance(self):
        voltage, current = self.read_voltage_and_current()
        return voltage

    def switch_front(self):
        self.write(":ROUT:TERM FRON")

    def switch_rear(self):
        self.write(":ROUT:TERM REAR")



# Written by Adam McCaughan Jan 17, 2014
# Run add_path.py first

import os
from datetime import datetime

# import cPickle as pickle
import matplotlib.cm as cm
import numpy as np
import scipy.io
from matplotlib import pyplot as plt


def save_xy_vs_param(
    X,
    Y,
    P,
    xname="x",
    yname="y",
    pname="p",
    test_type="Y Measurement",
    test_name="Test01",
    xlabel="amps",
    ylabel="volts",
    plabel="farads",
    title="",
    legend=False,
    xscale=1,
    yscale=1,
    pscale=1,
    comments="",
    filedir="",
    display_plot=False,
    zip_file=True,
):
    """Save list-of-lists X and Y data, along with list P (parameter) as MATLAB .mat file and Python .pickle file
    and additionally zips those .mat and .pickle files together with a .png of the graph
    """
    X = np.array(X)
    Y = np.array(Y)
    P = np.array(P)
    data_dict = {xname: X, yname: Y, pname: P, "comments": comments}
    time_str = datetime.now().strftime("%Y-%m-%d %H-%M-%S")
    file_name = "%s %s %s" % (test_type, time_str, test_name)
    file_path = os.path.join(filedir, file_name)

    # Save data in MATLAB and pickle formats
    scipy.io.savemat(file_path + ".mat", mdict=data_dict)
    # f = open(file_path + '.pickle', 'wb'); pickle.dump(data_dict, f); f.close()

    # Plot data
    if X.ndim == 1:
        plt.plot(X * xscale, Y * yscale)
    elif X.ndim == 2:
        mycm = cm.winter
        cm_indices = np.linspace(0, mycm.N * 0.8, len(X))
        for n in range(len(X)):
            plt.plot(
                X[n] * xscale,
                Y[n] * yscale,
                label="%0.3f %s" % (P[n] * pscale, plabel),
                c=mycm(int(cm_indices[n])),
            )
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    if legend is True:
        plt.legend(loc="best")
    plt.gca().get_xaxis().get_major_formatter().set_useOffset(
        False
    )  # Remvoes relative shift on graph
    plt.savefig(file_path + ".png")

    print("Saving data and figure with filename: %s" % file_name)
    # zf = zipfile.ZipFile(file_path + '.zip', 'w')
    # for name in ['.pickle', '.mat', '.png']:
    #    zf.write(file_path + name, arcname = file_name + name, compress_type = zipfile.ZIP_DEFLATED)

    if display_plot is True:
        plt.show()
    else:
        plt.close()
    return file_path


def save_x_vs_param(
    X,
    P,
    xname="x",
    pname="p",
    test_type="X Measurement",
    test_name="Test01",
    comments="",
    filedir="",
    zip_file=True,
):
    """Save list-of-lists X data, along with list P (parameter) as MATLAB .mat file and Python .pickle file
    and additionally zips those .mat and .pickle files together with a .png of the graph
    """
    data_dict = {xname: X, pname: P, "comments": comments}
    time_str = datetime.now().strftime("%Y-%m-%d %H-%M-%S")
    file_name = "%s %s %s" % (test_type, time_str, test_name)
    file_path = os.path.join(filedir, file_name)

    # Save data in MATLAB and pickle formats
    scipy.io.savemat(file_path + ".mat", mdict=data_dict)
    # f = open(file_path + '.pickle', 'wb'); pickle.dump(data_dict, f); f.close()

    print("Saving data and figure with filename: %s" % file_name)
    # zf = zipfile.ZipFile(file_path + '.zip', 'w')
    # for name in ['.pickle', '.mat']:
    #    zf.write(file_path + name, arcname = file_name + name, compress_type = zipfile.ZIP_DEFLATED)

    return (file_path, file_name)


def save_data_dict(
    data_dict, test_type="X Measurement", test_name="Test01", filedir="", zip_file=True
):
    """Directly input the data dictionary that scipy.io.savemat wants, and automatically zip the pickle
    and matlab file up"""
    time_str = datetime.now().strftime("%Y-%m-%d %H-%M-%S")
    file_name = "%s %s %s" % (test_type, time_str, test_name)
    file_path = os.path.join(filedir, file_name)

    # Save data in MATLAB and pickle formats
    scipy.io.savemat(file_path + ".mat", mdict=data_dict)
    # f = open(file_path + '.pickle', 'wb'); pickle.dump(data_dict, f); f.close()

    print("Saving data and figure with filename: %s" % file_name)
    # zf = zipfile.ZipFile(file_path + '.zip', 'w')
    # for name in ['.mat']:
    #     # zf.write(file_path + name, arcname = file_name + name, compress_type = zipfile.ZIP_DEFLATED)
    #     zf.write(file_path + name, arcname = file_name + name)

    return (file_path, file_name)


# X = []
# Y = []
# P = []
# for n in range(10):
#     X.append(np.linspace(0,10,100))
#     Y.append(np.random.rand(100) + n)
#     P.append(n)


# fn = save_xy_vs_param(X, Y, P, xname = 'f', yname = 'output', pname = 'shift',
#                         test_type = 'ZZZ Measurement', test_name = 'Test01',
#                         xlabel = 'Frequency (GHz)', ylabel = 'Whatup', title = 'Hey there',
#                         filedir = '/Users/amcc/Documents/MATLAB',
#                         legend = True, display_plot = True)



import time

import numpy as np
import pyvisa


class Agilent53230a(object):
    """Python class for Agilent 53230a counter, written by Adam McCaughan,
    modified from 5131a for 53230a by Emma Batson
    Use like c = Agilent53131a('GPIB0::3')"""

    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def reset(self):
        self.write("*RST")

    def basic_setup(self):
        self.write("*RST")
        self.write("*CLS")

        self.write("INP:LEV:AUTO OFF")  # Turn off auto trigger level
        self.write("INP:LEV -0.200V")  # Set trigger level
        self.write("INP:SLOP POS")  # Or POS. Trigger on negative slope
        self.write("INP:HYST:REL 0")  # Set hysteresis 0 indicates high sensitivity
        self.write("INP:COUP AC")  # Or DC.  Input coupling
        # self.write(':INP:IMP 50') # Set input impedance to 50ohms
        self.write("INP:IMP 1E6")  # Set input impedance to 1 MOhms

        self.write("INP:FILT OFF")  # Turn off 100kHz lowpass filter
        self.write('FUNC "TOT 1"')  # Totalize on channel 1
        self.write(
            "TOT:ARM:STAR:SOUR IMM"
        )  # Set start source to immediate (run on command)
        self.write(
            "TOT:ARM:STOP:SOUR TIM"
        )  # Set stop source to time (wait certain time)
        self.write("TOT:ARM:STOP:TIM 0.1")  # Set stop time to 100 ms
        self.write("INP:ATT 1")  # Or 10. Set attenuation factor

    def set_impedance(self, value=0):
        """
        VALUE=0 : IMPEDANCE IS 50OHMS
        VALUE=1 : IMPEDANCE IS 1MOHM

        """
        if value == 0:
            self.write(":INP:IMP 50")
        else:
            self.write(":INP:IMP 1E6")

    def set_trigger(self, trigger_voltage=-0.075, trigger_slope=None):
        if trigger_slope == "POS" or trigger_slope == "NEG":
            self.write(
                ":EVEN:SLOP %s" % trigger_slope
            )  # Or POS. Trigger on negative slope
        self.write(":EVEN:LEV %0.3fV" % trigger_voltage)  # Set trigger level

    def count_rate(self, counting_time=0.1):
        self.write(
            ":TOT:ARM:STOP:TIM %0.3f" % counting_time
        )  # Set stop time to # of seconds
        dcr = self.query(":READ?")
        dcr = float(dcr) / counting_time
        # time.sleep(counting_time + 0.1)
        return dcr

    def counts_vs_time(self, trigger_voltage=-0.075, counting_time=0.1, total_time=2):
        self.set_trigger(trigger_voltage)
        num_tests = int(total_time / counting_time)
        dcr = []
        t = []
        start_time = time.time()
        for n in range(num_tests):
            dcr.append(self.get_dcr(counting_time))
            t.append(time.time() - start_time)

        return t, dcr

    def scan_trigger_voltage(
        self, voltage_range=[-0.2, 0.2], counting_time=0.1, num_pts=40
    ):
        v = np.linspace(voltage_range[0], voltage_range[1], num_pts)
        dcr = []
        for trigger_voltage in v:
            self.set_trigger(trigger_voltage)
            dcr.append(self.count_rate(counting_time))
            print(
                "Trigger voltage = %0.3f  /  Count rate %0.1f"
                % (trigger_voltage, dcr[-1])
            )
        return v, np.array(dcr)



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 27 12:21:00 2019

@author: dizhu
"""

# %%
import sys

import matplotlib.pyplot as plt
import pyvisa

sys.path.append(r"C:\Users\dizhu\Desktop\GitHub\qnn-lab-instr-python")

from qnnpy.instruments.yokogawa_aq6370 import AQ6370

rm = pyvisa.ResourceManager()
rm.list_resources()


osa = AQ6370("GPIB0::7::INSTR")

# %%


osa.quick_setup()

wl, power = osa.single_sweep()
plt.plot(wl * 1e9, power)


# %%

# repeat_sweep(osa)

# %%

# %%

idn(osa)



import numpy as np
import pyvisa


# the commonds can be found in the following link
# http://rfmw.em.keysight.com/bihelpfiles/Trueform/webhelp/US/Default.htm?lc=eng&cc=US&id=2197433
class Agilent33600a(object):
    """Python class for Agilent 33600a 80MHz Frequency Generator, written by Adam McCaughan
    Modified for qnnpy compatibility by Emma Batson
    """

    def __init__(self, visa_name):
        rm = pyvisa.ResourceManager()
        self.pyvisa = rm.open_resource(visa_name)
        self.pyvisa.timeout = 5000  # Set response timeout (in milliseconds)
        # self.pyvisa.query_delay = 1 # Set extra delay time between write and read commands

    def read(self):
        return self.pyvisa.read()

    def write(self, string):
        self.pyvisa.write(string)

    def query(self, string):
        return self.pyvisa.query(string)

    def beep(self):
        self.pyvisa.write("SYST:BEEP")

    def reset(self):
        self.write("*RST")

    def sync_phase(self, ch="Ch1"):
        if ch == "Ch1":
            channel = 1
        else:
            channel = 2
        self.write("SOURce{}:PHASe:SYNChronize".format(channel))

    def set_usual_unit(self):
        self.write("SOURce1:APPLy?")
        self.write("DISP:UNIT:ARBR FREQ")
        self.write("DISP:UNIT:RATE FREQ")
        self.write("DISP:UNIT:VOLT HIGH")
        self.write("DISP:VIEW STAN")

    def set_dual_channel_coupling(self, ch="Ch1", state="ON"):
        if state == "ON":
            if ch == "Ch1":
                self.write("SOUR1:FREQ:COUP ON")
            elif ch == "Ch2":
                self.write("SOUR2:FREQ:COUP ON")
        elif state == "OFF":
            if ch == "Ch1":
                self.write("SOUR1:FREQ:COUP OFF")
            elif ch == "Ch2":
                self.write("SOUR2:FREQ:COUP OFF")
        else:
            print("error")

    def set_sin(self, freq=1000, vpp=0.1, voffset=0):
        # In a string, %0.6e converts a number to scientific notation like
        # print '%.6e' %(1234.56789) outputs '1.234568e+03'
        self.write("APPL:SIN %0.6e HZ, %0.6e VPP, %0.6e V" % (freq, vpp, voffset))

    def set_square(
        self, ch="Ch1", freq=1000, vpp=0.1, voffset=0, duty_cycle=0.5, phase=0.0
    ):
        # duty cycle limit
        # 20% to 80% (frequency < 25 MHz)
        # 40% to 60% (25 MHz < frequency < 50 MHz)
        # 50% (frequency > 50 MHz)
        if ch == "Ch1":
            channel = 1
        elif ch == "Ch2":
            channel = 2

        self.write(
            "SOUR%1.0d:APPLy:SQUare %0.6e Hz, %0.6e VPP, %.6e V"
            % (channel, freq, vpp, voffset)
        )
        self.write(
            "SOUR%1.0d:FUNCtion:SQUare:DCYCle %0.2e" % (channel, duty_cycle * 100)
        )
        self.write("SOUR%1.0d:PHAS %0.3e" % (channel, phase))

    def set_ramp(self, ch="Ch1", freq=1000, vpp=0.1, voffset=0, symm=100):
        if ch == "Ch1":
            channel = 1
        elif ch == "Ch2":
            channel = 2
        # Ramp frequency is limited to 200 kHz
        self.write("SOUR{}:APPL:RAMP {}, {}, {}".format(channel, freq, vpp, voffset))
        self.write("SOUR{}:FUNC:RAMP:SYMM {}".format(channel, symm))

    def set_triangle(self, ch="Ch1", freq=1000, vpp=0.1, voffset=0, phase=0):
        if ch == "Ch1":
            channel = 1
        elif ch == "Ch2":
            channel = 2
        # Ramp frequency is limited to 200 kHz
        self.write("SOUR{}:APPL:TRI {}, {}, {}".format(channel, freq, vpp, voffset))
        self.write("SOUR%1.0d:PHAS %0.3e" % (channel, phase))

    def set_dc(self, ch="Ch1", voffset=0):
        if ch == "Ch1":
            channel = 1
        elif ch == "Ch2":
            channel = 2

        self.write("SOUR%1.0d:APPL:DC" % (channel))
        self.write("SOUR%1.0d:VOLT:OFFS %0.3e" % (channel, voffset))

    def set_pulse(
        self,
        ch="Ch1",
        freq=1000,
        vlow=0.0,
        vhigh=1.0,
        phase=1.0,
        width=100e-6,
        leading_edge_time=2.9e-9,
        trailing_edge_time=2.9e-9,
    ):
        if ch == "Ch1":
            channel = 1
        elif ch == "Ch2":
            channel = 2
        else:
            print("error")

        vpp = vhigh - vlow
        voffset = (vhigh + vlow) / 2.0

        self.write("SOUR%1.0d:APPL:PULS %d,%0.6e,%0.6e" % (channel, freq, vpp, voffset))
        self.write("SOUR%1.0d:FUNC:PULS:WIDT %0.6e" % (channel, width))
        self.write("SOUR%1.0d:FUNC:PULS:TRAN:LEAD %0.6e" % (channel, leading_edge_time))
        self.write("SOUR%1.0d:FUNC:PULS:TRAN:TRA %0.6e" % (channel, trailing_edge_time))
        self.write("SOUR%1.0d:PHAS %0.3e" % (channel, phase))

    def set_freq(self, freq=1000):
        self.write("FREQ %0.6e" % (freq))

    def set_vpp(self, vpp=0.1):
        self.write("VOLT %0.6e" % (vpp))

    def set_voffset(self, voffset=0.0):
        self.write("VOLT:OFFS %0.6e" % (voffset))

    def set_phase(self, ch="Ch1", angle=0):
        if ch == "Ch1":
            channel = 1
        elif ch == "Ch2":
            channel = 2
        self.write("SOURce{}:PHASe {}".format(channel, angle))

    def set_vhighlow(self, vlow=0.0, vhigh=1.0):
        if vhigh > vlow:
            self.set_vpp(vhigh - vlow)
            self.set_voffset((vhigh + vlow) / 2.0)
            self.set_polarity(inverted=False)
        elif vhigh < vlow:
            self.set_vpp(vlow - vhigh)
            self.set_voffset((vhigh + vlow) / 2.0)
            self.set_polarity(inverted=True)

    def set_output(self, output=False, ch="both"):
        """breaking legacy behavior to make compatible with
        other sources. default behavior is to turn both channels
        on or off
        """
        if output:
            state = "ON"
        else:
            state = "OFF"
        if ch == "both" or ch == "Ch1":
            self.write("OUTP1 {}".format(state))
        if ch == "both" or ch == "Ch2":
            self.write("OUTP2 {}".format(state))

    def set_output_legacy(self, ch="Ch1", output="ON"):
        if ch == "Ch1":
            channel = 1
        elif ch == "Ch2":
            channel = 2

        if output in ["on", "ON", "On", "oN"]:
            self.write("OUTP%1.0d ON" % (channel))
        else:
            self.write("OUTP%1.0d OFF" % (channel))

    def set_sync(self, output="ON"):
        self.write("OUTP:SYNC:MODE NORM")
        self.write("OUTP:SYNC:POL NORM")
        if output in ["on", "ON", "On", "oN"]:
            self.write("OUTP:SYNC ON")
        else:
            self.write("OUTP:SYNC OFF")

    def set_load(self, high_z=False):
        if high_z is True:
            self.write("OUTP:LOAD INF")
        else:
            self.write("OUTP:LOAD 50")

    def set_polarity(self, inverted=False):
        if inverted is True:
            self.write("OUTP:POL INV")
        else:
            self.write("OUTP:POL NORM")

    def set_trigger(self, external_trigger=False, delay=0.0):
        if external_trigger:
            self.write("TRIG:SOUR EXT")
        else:
            self.write("TRIG:SOUR IMM")
        self.write("TRIG:DEL %s" % (delay))  # Delay in seconds

    def trigger_now(self):
        self.write("*TRG")

    def set_burst_mode(self, ch=1, burst_enable=True, num_cycles=1, phase=0):
        if burst_enable:
            self.write("SOUR%1.0d:BURS:STAT ON" % (ch))  # Enables burst state
            self.write("SOUR%1.0d:BURS:NCYC %s" % (ch, num_cycles))
            self.write("SOUR%1.0d:BURS:PHAS %s" % (ch, phase))  # Phase in degrees

        else:
            self.write("SOUR%1.0d:BURS:STAT OFF" % (ch))  # Disables burst state

    def set_arb_wf(self, t=[0.0, 1e-3], v=[0.0, 1.0], name="ARB_PY"):
        """Input voltage values will be scaled to +/-1.0, you can then adjust the overall
        amplitude using the set_vpp function.  The 33250a does not allow the input of time for each
        point, so we instead use interpolation here to create waveform of 2^14 equally-spaced
        points, after which you can use set_freq to get the desired freq"""

        t = np.array(t)
        v = np.array(v)

        v = v - min(v)
        v = 2 * v / max(v)
        v = v - 1
        temp = self.timeout
        self.timeout = 60
        t_interp = np.linspace(t[0], t[-1], 2**14)  # Can be up to 2**14 long
        v_interp = np.interp(t_interp, t, v)

        data_strings = ["%0.3f" % x for x in v_interp]
        data_msg = ", ".join(data_strings)

        self.write(
            "DATA VOLATILE, " + data_msg
        )  # Form of "DATA VOLATILE, 1, .67, .33, 0, -.33", p200 user's guide
        name = name[0:8].upper()
        self.write("DATA:COPY %s, VOLATILE" % name)
        self.write("APPL:USER")  # Set output to ARB
        self.write("FUNC:USER %s" % name)  # Select the waveform in the volatile memory
        self.write("APPL:USER")
        self.timeout = temp
        # self.write('FUNC USER') # Output the selected waveform

    """
    def setup_heartbeat_wf(self):
        heartbeat_t = [0.0, 4.0/8, 5.0/8, 6.0/8,  7.0/8, 8.0/8]
        heartbeat_v = [0.0,   0.0,   1.0,   0.0,   -1.0,   0.0]
        freq_gen.set_arb_wf(t = heartbeat_t, v = heartbeat_v, name = 'HEARTBEA')
    """

    def select_arb_wf(self, name="HEARTBEA"):
        name = name[0:8].upper()
        self.write("APPL:USER")  # Set output to ARB
        self.write("FUNC:USER %s" % name)
        self.write("APPL:USER")  # Set output to ARB
